import os
import re
import sys
import platform
import subprocess

from setuptools import Extension
from distutils.version import LooseVersion
from setuptools.command.build_ext import build_ext


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: " +
                ", ".join(e.name for e in self.extensions))

        if platform.system() == "Windows":
            cmake_version = LooseVersion(re.search(r'version\s*([\d.]+)',
                                         out.decode()).group(1))
            if cmake_version < '3.1.0':
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(
            os.path.dirname(self.get_ext_fullpath(ext.name)))

        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable]

        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        if platform.system() == "Windows":
            cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(
                cfg.upper(),
                extdir)]
            if sys.maxsize > 2**32:
                cmake_args += ['-A', 'x64']
            build_args += ['--', '/m']
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            build_args += ['--', '-j2']

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(
            env.get('CXXFLAGS', ''),
            self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args,
                              cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] + build_args,
                              cwd=self.build_temp)


def check_env():
    NOISEFLOW_USE_CPP = (os.getenv('NOISEFLOW_USE_CPP', '') == '1')

    ext_modules = []
    if NOISEFLOW_USE_CPP:
        ext_modules.append(CMakeExtension('noiseflow/lib/'))

    return ext_modules


def build(setup_kwargs):
    """
    This is a callback for poetry used to hook in our extensions.
    """
    setup_kwargs.update(
        {
            "ext_modules": check_env(),
            "cmdclass": {"build_ext": CMakeBuild},
        }
    )



# The stack functions are from https://github.com/xtyangpsp/StackMaster

import time
import numpy as np
import matplotlib.pyplot as plt

from tqdm import tqdm 
from joblib import Parallel, delayed
from itertools import compress
from scipy.signal import hilbert
from scipy.fftpack import fft,ifft,next_fast_len
from noiseflow.cc.python.utils import slice_window, split

try:
    from stockwell import st
    from tslearn.utils import to_time_series_dataset
    from tslearn.clustering import TimeSeriesKMeans
except:
    pass



def rms(d):
    return np.sqrt(np.mean(d**2))


def power2pad(data):
	"""Zero pad data such that its length is a power of 2"""
	N=int(2**np.ceil(np.log2(len(data))))
	pad_end=np.zeros(int(N-len(data)))

	return np.concatenate((data,pad_end))


def stack(d,method,par=None):
    """
    this is a wrapper for calling individual stacking functions.
    d: data. 2-d array
    method: stacking method, one of "linear","pws","robust","acf","nroot","selective",
            "cluster"
    par: dictionary containing all parameters for each stacking method. defaults will
        be used if not specified.

    RETURNS:
    ds: stacked data, which may be a list depending on the method.
    """
    method_list=["linear","pws","robust","acf","nroot","selective",
            "cluster","tfpws"]
    if method not in method_list:
        raise ValueError("$s not recoganized. use one of $s"%(method,str(method_list)))
    
    par0={"axis":0,"p":2,"g":1,"cc_min":0.0,"epsilon":1E-5,"maxstep":10,
            "win":None,"stat":False,"h":0.75,'plot':False,'normalize':True,'ref':None}  #stat: if true, will return statistics.
    
    if par is None:
        par=par0
    else:
        par={**par0,**par} #use par values if specified. otherwise, use defaults.

    if method.lower() == 'linear':
        ds = np.mean(d,axis=par["axis"])
    elif method.lower() == 'pws':
        ds = pws(d,p=par['p'])
    elif method.lower() == 'tfpws':
        ds = tfpws(d,p=par['p'])
    # elif method.lower() == 'tfpws-dost':
    #     ds = tfpws_dost(d,p=par['p'])
    elif method.lower() == 'robust':
        ds = robust(d,epsilon=par['epsilon'],maxstep=par['maxstep'],win=par["win"],
                stat=par['stat'],ref=par['ref'])
    elif method.lower() == 'acf':
        ds = adaptive_filter(d,g=par['g'])
    elif method.lower() == 'nroot':
        ds = nroot(d,p=par['p'])
    elif method.lower() == 'selective':
        ds = selective(d,cc_min=par['cc_min'],epsilon=par['epsilon'],maxstep=par['maxstep'],
                stat=par['stat'],ref=par['ref'],win=par["win"])
    elif method.lower() == 'cluster':
        ds = clusterstack(d,h=par['h'],axis=par['axis'],win=par["win"],
        normalize=par['normalize'],plot=par['plot'])
    #
    return ds

def seisstack(d,method,par=None):
    """
    This is the same as stack(), to be compatible with old usage.
    """
    return stack(d,method=method,par=par)

def robust(d,epsilon=1E-5,maxstep=10,win=None,stat=False,ref=None):
    """
    this is a robust stacking algorithm described in Pavlis and Vernon 2010. Generalized
    by Xiaotao Yang.

    PARAMETERS:
    ----------------------
    d: numpy.ndarray contains the 2D cross correlation matrix
    epsilon: residual threhold to quit the iteration (a small number). Default 1E-5
    maxstep: maximum iterations. default 10.
    win: [start_index,end_index] used to compute the weight, instead of the entire trace. Default None.
            When None, use the entire trace.
    ref: reference stack, with the same length as individual data. Default: None. Use median().
    RETURNS:
    ----------------------
    newstack: numpy vector contains the stacked cross correlation

    Written by Marine Denolle
    Modified by Xiaotao Yang
    """
    if d.ndim == 1:
        print('2D matrix is needed')
        return d
    N,M = d.shape
    res  = 9E9  # residuals
    w = np.ones(d.shape[0])
    small_number=1E-15
    nstep=0
    if N >=2:
        if ref is None:
            newstack = np.median(d,axis=0)
        else:
            newstack = ref
        if win is None:
            win=[0,-1]
        while res > epsilon and nstep <=maxstep:
            stack = newstack
            for i in range(d.shape[0]):
                dtemp=d[i,win[0]:win[1]]
                crap = np.multiply(stack[win[0]:win[1]],dtemp.T)
                crap_dot = np.sum(crap)
                di_norm = np.linalg.norm(dtemp)
                ri_norm = np.linalg.norm(dtemp -  crap_dot*stack[win[0]:win[1]])
                if ri_norm < small_number:
                    w[i]=0
                else:
                    w[i]  = np.abs(crap_dot) /di_norm/ri_norm
            w =w /np.sum(w)
            newstack =np.sum( (w*d.T).T,axis=0)#/len(cc_array[:,1])
            res = np.linalg.norm(newstack-stack,ord=1)/np.linalg.norm(newstack)/len(d[:,1])
            nstep +=1
    else:
        newstack=d[0].copy()
    if stat:
        return newstack, w, nstep
    else:
        return newstack

def adaptive_filter(d,g=1):
    '''
    the adaptive covariance filter to enhance coherent signals. Fellows the method of
    Nakata et al., 2015 (Appendix B)

    the filtered signal [x1] is given by x1 = ifft(P*x1(w)) where x1 is the ffted spectra
    and P is the filter. P is constructed by using the temporal covariance matrix.

    PARAMETERS:
    ----------------------
    d: numpy.ndarray contains the 2D traces of daily/hourly cross-correlation functions
    g: a positive number to adjust the filter harshness [default is 1]
    RETURNS:
    ----------------------
    newstack: numpy vector contains the stacked cross correlation function
    '''
    if d.ndim == 1:
        print('2D matrix is needed')
        return d
    N,M = d.shape
    if N>=2:
        Nfft = next_fast_len(M)

        # fft the 2D array
        spec = fft(d,axis=1,n=Nfft)[:,:M]

        # make cross-spectrm matrix
        cspec = np.zeros(shape=(N*N,M),dtype=np.complex64)
        for ii in range(N):
            for jj in range(N):
                kk = ii*N+jj
                cspec[kk] = spec[ii]*np.conjugate(spec[jj])

        S1 = np.zeros(M,dtype=np.complex64)
        S2 = np.zeros(M,dtype=np.complex64)
        # construct the filter P
        for ii in range(N):
            mm = ii*N+ii
            S2 += cspec[mm]
            for jj in range(N):
                kk = ii*N+jj
                S1 += cspec[kk]

        p = np.power((S1-S2)/(S2*(N-1)),g)

        # make ifft
        narr = np.real(ifft(np.multiply(p,spec),Nfft,axis=1)[:,:M])
        newstack=np.mean(narr,axis=0)
    else:
        newstack=d[0].copy()
    #
    return newstack

def pws(d,p=2):
    '''
    Performs phase-weighted stack on array of time series. Modified on the noise function by Tim Climents.
    Follows methods of Schimmel and Paulssen, 1997.
    If s(t) is time series data (seismogram, or cross-correlation),
    S(t) = s(t) + i*H(s(t)), where H(s(t)) is Hilbert transform of s(t)
    S(t) = s(t) + i*H(s(t)) = A(t)*exp(i*phi(t)), where
    A(t) is envelope of s(t) and phi(t) is phase of s(t)
    Phase-weighted stack, g(t), is then:
    g(t) = 1/N sum j = 1:N s_j(t) * | 1/N sum k = 1:N exp[i * phi_k(t)]|^v
    where N is number of traces used, v is sharpness of phase-weighted stack

    PARAMETERS:
    ---------------------
    d: N length array of time series data (numpy.ndarray)
    p: exponent for phase stack (int). default is 2

    RETURNS:
    ---------------------
    newstack: Phase weighted stack of time series data (numpy.ndarray)
    '''

    if d.ndim == 1:
        print('2D matrix is needed')
        return d
    N,M = d.shape
    if N >=2:
        analytic = hilbert(d,axis=1, N=next_fast_len(M))[:,:M]
        phase = np.angle(analytic)
        phase_stack = np.mean(np.exp(1j*phase),axis=0)
        phase_stack = np.abs(phase_stack)**(p)

        weighted = np.multiply(d,phase_stack)

        newstack=np.mean(weighted,axis=0)
    else:
        newstack=d[0].copy()
    return newstack

def nroot(d,p=2):
    '''
    this is nth-root stacking algorithm translated based on the matlab function
    from https://github.com/xtyangpsp/SeisStack (by Xiaotao Yang; follows the
    reference of Millet, F et al., 2019 JGR)

    Parameters:
    ------------
    d: numpy.ndarray contains the 2D cross correlation matrix
    p: np.int, nth root for the stacking. Default is 2.

    Returns:
    ------------
    newstack: np.ndarray, final stacked waveforms

    Written by Chengxin Jiang @ANU (May2020)
    '''
    if d.ndim == 1:
        print('2D matrix is needed for nroot_stack')
        return d
    N,M = d.shape
    if N >=2:
        dout = np.zeros(M,dtype=np.float32)

        # construct y
        for ii in range(N):
            dat = d[ii,:]
            dout += np.sign(dat)*np.abs(dat)**(1/p)
        dout /= N

        # the final stacked waveform
        newstack = dout*np.abs(dout)**(p-1)
    else:
        newstack=d[0].copy()

    return newstack


def selective(d,cc_min,epsilon=1E-5,maxstep=10,win=None,stat=False,ref=None):
    '''
    this is a selective stacking algorithm developed by Jared Bryan/Kurama Okubo.

    PARAMETERS:
    ----------------------
    d: numpy.ndarray contains the 2D cross correlation matrix
    epsilon: residual threhold to quit the iteration
    cc_min: numpy.float, threshold of correlation coefficient to be selected
    epsilon: residual threhold to quit the iteration (a small number). Default 1E-5
    maxstep: maximum iterations. default 10.
    win: [start_index,end_index] used to compute the weight, instead of the entire trace. Default None.
            When None, use the entire trace.
    ref: reference stack, with the same length as individual data. Default: None. Use mean().
    RETURNS:
    ----------------------
    newstack: numpy vector contains the stacked cross correlation
    nstep: np.int, total number of iterations for the stacking

    Originally ritten by Marine Denolle
    Modified by Chengxin Jiang @Harvard (Oct2020)
    '''
    if d.ndim == 1:
        print('2D matrix is needed for selective stacking')
        return d
    N,M = d.shape
    if N>=2:
        res  = 9E9  # residuals
        cof  = np.zeros(N,dtype=np.float32)
        if ref is None:
            newstack = np.mean(d,axis=0)
        else:
            newstack = ref

        nstep = 0
        if win is None:
            win=[0,-1]
        # start iteration
        while res>epsilon and nstep<=maxstep:
            for ii in range(N):
                cof[ii] = np.corrcoef(newstack[win[0]:win[1]], d[ii,win[0]:win[1]])[0, 1]

            # find good waveforms
            indx = np.where(cof>=cc_min)[0]
            nstep +=1
            if not len(indx):
                newstack=np.ndarray((d.shape[1],))
                newstack.fill(np.nan)
                print('cannot find good waveforms inside selective stacking')
                break
            else:
                oldstack = newstack
                newstack = np.mean(d[indx],axis=0)
                res = np.linalg.norm(newstack-oldstack)/(np.linalg.norm(newstack)*M)
    else:
        newstack=d[0].copy()
    if stat:
        return newstack, nstep
    else:
        return newstack
#
def clusterstack(d,h=0.75,win=None,axis=0,normalize=True,plot=False):
    '''
    Performs stack after clustering. The data will be clustered into two groups.
    If the two centers of the clusters are similar (defined by corrcoef >= "t"), the original
    traces associated with both clusters will be used to produce the final linear stack, weighted by
    normalized SNR (phase clarity) of each cluster. Otherwise, the one with larger phase clarity
    (defined as max(abs(amplitudes))/rms(abs(amplitudes))) will be used to get the final stack.

    PARAMETERS:
    ---------------------
    d: N length array of time series data (numpy.ndarray)
    h: corrcoeff threshold to decide which group/cluster to use. Default 0.75.
    win: [start_index,end_index] used to compute the weight, instead of the entire trace. Default None.
            When None, use the entire trace.
    axis: which axis to stack. default 0.
    normalize: Normalize the traces before clustering. This will only influence the cluster.
            The final stack will be produced using the original data.
    plot: plot clustering results. default False.

    RETURNS:
    ---------------------
    newstack: final stack.
    '''
    ncluster=2 #DO NOT change this value.
    min_trace=2 #minimum of two traces.
    metric="euclidean" #matric to compute the distance in kmeans clustering.
    if d.ndim == 1:
        print('2D matrix is needed')
        return d
    N,M = d.shape
    if N >= min_trace:
        dataN=d.copy()
        if normalize:
            for i in range(N):
                dataN[i]=d[i]/np.max(np.abs(d[i]),axis=0)

        ts = to_time_series_dataset(dataN)

        km = TimeSeriesKMeans(n_clusters=ncluster, n_jobs=1,metric=metric, verbose=False,
                              max_iter_barycenter=100, random_state=0)
        y_pred = km.fit_predict(ts)
        snr_all=[]
        centers_all=[]
        cidx=[]
        if win is None:
            win=[0,-1]
        for yi in range(ncluster):
            cidx.append(np.where((y_pred==yi))[0])
            center=km.cluster_centers_[yi].ravel()#np.squeeze(np.mean(ts[y_pred == yi].T,axis=2))
            centers_all.append(center)
            snr=np.max(np.abs(center[win[0]:win[1]]))/rms(np.abs(center))
            snr_all.append(snr)

        #
        if plot:
            plt.figure(figsize=(12,4))
            for yi in range(ncluster):
                plt.subplot(1,ncluster,yi+1)
                plt.plot(np.squeeze(ts[cidx[yi]].T),'k-',alpha=0.3)
                plt.plot(centers_all[yi],'r-')
                plt.title('Cluster %d: %d'%(yi+1,len(cidx[yi])))
            plt.show()
        cc=np.corrcoef(centers_all[0],centers_all[1])[0,1]
        if cc>= h: #use all data
            snr_normalize=snr_all/np.sum(snr_all)
            newstack=np.zeros((M))
            for yi in range(ncluster):
                newstack += snr_normalize[yi]*np.mean(d[cidx[yi]],axis=0)
        else:
            goodidx=np.argmax(snr_all)
            newstack=np.mean(d[cidx[goodidx]],axis=0)
        del dataN,ts,y_pred
    else:
        newstack=d[0].copy()
    #
    return newstack

def tfpws(d,p=2,axis=0):
    '''
    Performs time-frequency domain phase-weighted stack on array of time series.

    $C_{ps} = |(\sum{S*e^{i2\pi}/|S|})/M|^p$, where $C_{ps}$ is the phase weight. Then
    $S_{pws} = C_{ps}*S_{ls}$, where $S_{ls}$ is the S transform of the linea stack
    of the whole data.

    Reference:
    Schimmel, M., Stutzmann, E., & Gallart, J. (2011). Using instantaneous phase
    coherence for signal extraction from ambient noise data at a local to a
    global scale. Geophysical Journal International, 184(1), 494–506.
    https://doi.org/10.1111/j.1365-246X.2010.04861.x

    PARAMETERS:
    ---------------------
    d: N length array of time series data (numpy.ndarray)
    p: exponent for phase stack (int). default is 2
    axis: axis to stack, default is 0.

    RETURNS:
    ---------------------
    newstack: Phase weighted stack of time series data (numpy.ndarray)
    '''
    if d.ndim == 1:
        print('2D matrix is needed')
        return d
    N,M = d.shape
    if N >=2:
        lstack=np.mean(d,axis=axis)
        #get the ST of the linear stack first
        stock_ls=st.st(power2pad(lstack))

        #run a ST to get the dimension of ST result
        stock_temp=st.st(power2pad(d[0]))
        phase_stack=np.zeros((stock_temp.shape[0],stock_temp.shape[1]),dtype='complex128')
        for i in range(N):
            if i>0: #zero index has been computed.
                stock_temp=st.st(power2pad(d[i]))
            phase_stack += np.multiply(stock_temp,np.angle(stock_temp))/np.abs(stock_temp)
        #
        phase_stack = np.abs(phase_stack/N)**p

        pwstock=np.multiply(phase_stack,stock_ls)
        recdostIn=np.real(st.ist(pwstock))
        newstack=recdostIn[:M] # trim padding
    else:
        newstack=d[0].copy()
    #
    return newstack

# def tfpws_dost(d,p=2,axis=0):
#     '''
#     Performs time-frequency domain phase-weighted stack on array of time series using DOST (Discrete
#     orthogonal stockwell transform).
#     $C_{ps} = |(\sum{S*e^{i2\pi}/|S|})/M|^p$, where $C_{ps}$ is the phase weight. Then
# 	$S_{pws} = C_{ps}*S_{ls}$, where $S_{ls}$ is the Discrete Orthonormal S transform
# 	of the linear stack of the whole data.

#     DOST stacking was implemented by Jared Bryan.

#     Reference for tf-PWS:
#     Schimmel, M., Stutzmann, E., & Gallart, J. (2011). Using instantaneous phase
#     coherence for signal extraction from ambient noise data at a local to a
#     global scale. Geophysical Journal International, 184(1), 494–506.
#     https://doi.org/10.1111/j.1365-246X.2010.04861.x

#     Reference for DOST:
#     U. Battisti, L. Riba, "Window-dependent bases for efficient representations of the
# 	Stockwell transform", Applied and Computational Harmonic Analysis, 23 February 2015,
# 	http://dx.doi.org/10.1016/j.acha.2015.02.002.

#     PARAMETERS:
#     ---------------------
#     d: N length array of time series data (numpy.ndarray)
#     p: exponent for phase stack (int). default is 2
#     axis: axis to stack, default is 0.

#     RETURNS:
#     ---------------------
#     newstack: Phase weighted stack of time series data (numpy.ndarray)
#     '''
#     if d.ndim == 1:
#         print('2D matrix is needed')
#         return d
#     N,M = d.shape
#     if N >=2:
#         lstack=np.mean(d,axis=axis)
#         #get the dost of the linear stack first
#         stock_ls_dost=DOST(lstack) # initialize dost object
#         stock_ls=stock_ls_dost.dost(stock_ls_dost.data) # calculate the dost

#     	# calculate dost for first trace to know its shape
#         stock_dost=DOST(d[0])
#         stock_temp=stock_dost.dost(stock_dost.data)
#     	# initialize stack
#         phase_stack=np.zeros(len(stock_temp),dtype='complex128')
#     	# calculate the dost for each trace to be stacked
#         for i in range(d.shape[0]):
#         	if i>0: # zero index has been computed
#         	    stock_dost=DOST(d[i])
#         	    stock_temp=stock_dost.dost(stock_dost.data)
                    
#             phase_stack+=np.multiply(stock_temp,np.angle(stock_temp))/np.abs(stock_temp)

#         phase_stack = np.abs(phase_stack/N)**p

#         pwstock=np.multiply(phase_stack,stock_ls)
#         recdostIn = np.real(stock_dost.idost(pwstock))
#         newstack = recdostIn[:M] # trim padding
#     else:
#         newstack=d[0].copy()
#     #
#     return newstack

class DOST:
    def __init__(self, data):
        # make sure data length is a power of 2
        if np.ceil(np.log2(len(data)))==np.floor(np.log2(len(data))):
            # length of data already a power of 2
            self.data=data
        else:
            # pad data to nearest power of 2
            self.data=self.pad(data)

    def pad(self,data):
        """Zero pad data such that its length is a power of 2

        PARAMETERS:
        ---------------------
        data: array of time series data (numpy.ndarray)

        RETURNS:
        ---------------------
        data: zero-padded array of time series data (numpy.ndarray)
        """
        N=int(2**np.ceil(np.log2(len(data))))
        pad_end=np.zeros(int(N-len(data)))
        data=np.concatenate((data,pad_end))

        return data

    def fourier(self, d):
        """Normalized and centered fft

        PARAMETERS:
        ---------------------
        d: array of time series data (numpy.ndarray)

        RETURNS:
        ---------------------
        fftIn: array of frequency-domain data (numpy.ndarray)
        """
        fftIn=(1/np.sqrt(len(d))) * np.fft.fftshift(np.fft.fft(np.fft.ifftshift(d)))
        return fftIn

    def ifourier(self,d):
        """Normalized and centered ifft

        PARAMETERS:
        ---------------------
        d: array of frequency-domain data (numpy.ndarray)

        RETURNS:
        ---------------------
        ifftIn: array of time series data (numpy.ndarray)
        """
        ifftIn=np.sqrt(len(d)) * np.fft.fftshift(np.fft.ifft(np.fft.ifftshift(d)))
        return ifftIn

    def dostbw(self,D):
        """Calculate size of the DOST bandwidths

        PARAMETERS:
        ---------------------
        D: length of time-series data (int)

        RETURNS:
        ---------------------
        bw: list of DOST bandwidths
        """
        bw=[0]
        bw.extend(np.arange(np.log2(D)-2, -1e-9, -1))
        bw.extend([0])
        bw.extend(np.arange(0, np.log2(D)-2+1e-9))
        bw=2**np.array(bw)
        return bw

    def dost(self,d):
        """Discrete Orthonormal Stockwell Transform

        PARAMETERS:
        ---------------------
        d: array of time-series data (numpy.ndarray)

        RETURNS:
        ---------------------
        d_dost: array of DOST coefficients (numpy.ndarray)
        """
        d_dost=self.fourier(d)
        D=len(d)
        bw=self.dostbw(D)
        k=0
        # heuristically, this is a short-time fourier transform with a frequency-dependent bandwidth
        for i in bw:
            i=int(i)
            if i==1:
                k=k+i
            else:
                d_dost[k:k+i] = self.ifourier(d_dost[k:k+i])
                k=k+i
        return d_dost

    def idost(self,d):
        """Inverse Discrete Orthonormal Stockwell Transform

        PARAMETERS:
        ---------------------
        d: array of DOST coefficients (numpy.ndarray)

        RETURNS:
        ---------------------
        d_idost: array of time-series data (numpy.ndarray)
        """
        d_idost=d
        D=len(d)
        bw=self.dostbw(D)
        k=0
        for i in bw:
            i=int(i)
            if i==1:
                k=k+i
            else:
                d_idost[k:k+i] = self.fourier(d_idost[k:k+i])
                k=k+i
        d_idost = self.ifourier(d_idost)
        return d_idost


class StackClass_python(object):
    def __init__(self, corr_data, stack_method, par, stack_all, stack_len, stack_step, pick, median_high, median_low, flag, jobs):
        self.corr_data = corr_data
        self.stack_method = stack_method
        self.par = par
        self.stack_all = stack_all
        self.stack_len = stack_len
        self.stack_step = stack_step
        self.pick_m = pick
        self.median_high = median_high
        self.median_low = median_low
        self.flag = flag
        self.jobs = jobs

        self.pair_num = int(corr_data.shape[0])
        self.corr_win_num = int(corr_data.shape[1])
        self.npts = int(corr_data.shape[2])

        # slice_window --> _stack_win_num, _win_info
        if stack_all:
            self.win_info = np.array([[0, self.corr_win_num]])
            self.stack_win_num = 1
        else:
            self.win_info = slice_window(self.corr_win_num, stack_len, stack_step)
            self.stack_win_num = int(self.win_info.shape[0])

        # initialize output_data
        self.ngood_all = np.empty((self.pair_num, self.stack_win_num))
        if self.corr_data.dtype == np.dtype(np.float32):
            self.output_data = np.empty((self.pair_num, self.stack_win_num, self.npts), dtype=np.float32)
            self.time_dtype = np.float32
            self.freq_dtype = np.complex64
        elif self.corr_data.dtype == np.dtype(np.float64):
            self.output_data = np.empty((self.pair_num, self.stack_win_num, self.npts), dtype=np.float64)
            self.time_dtype = np.float64
            self.freq_dtype = np.complex128
        else:
            print("error: please input the correct data type.")
            exit(1)


    def pick(self, pair_id):
        nindex = None
        if self.pick_m:
            # SOME WARNING HERE!!!
            ampmax = np.amax(self.corr_data[pair_id, :, :], axis=1)
            median_ampmax = np.median(ampmax)
            index = np.where((ampmax > self.median_low * median_ampmax) & (ampmax < self.median_high * median_ampmax))[0]
        
            nindex = np.array(index, dtype=int)
        else:
            nindex = np.arange(0, self.corr_win_num, dtype=int)

        return nindex


    def check_ngood(self, nindex):
        ngood = np.empty(self.stack_win_num)
        if self.pick_m:
            for i in range(0, self.stack_win_num):
                ngood[i] = np.sum((nindex >= self.win_info[i, 0]) & (nindex < self.win_info[i, 1]))
        else:
            ngood = self.stack_len * np.ones(self.stack_win_num)

        return ngood


    def stack(self, pair_id, nindex):
        stack_data = np.empty((self.stack_win_num, self.npts), dtype=self.time_dtype)

        for i in range(0, self.stack_win_num):
            each_index = np.arange(self.win_info[i, 0], self.win_info[i, 1])
            pick_index = list(compress(each_index, np.isin(each_index, nindex)))

            if len(pick_index) == 0:
                stack_data[i, :] = np.zeros(self.npts)
            else:
                pick_corr_data = self.corr_data[pair_id, pick_index, :]
                stack_data[i, :] = stack(pick_corr_data, self.stack_method, self.par)    

        return stack_data


    def process_chunk(self, chunk_start, chunk_end):
        ngood_results = []
        stack_data_results = []

        if self.flag and chunk_start == 0:
            bar = tqdm(range(chunk_start, chunk_end))
        else:
            bar = range(chunk_start, chunk_end)

        for i in bar:
            # pick --> _nindex
            nindex = self.pick(i)

            # ngood
            ngood = self.check_ngood(nindex)

            # corr
            stack_data = self.stack(i, nindex) # "linear", "pws", "robust"

            ngood_results.append(ngood)
            stack_data_results.append(stack_data)

        return ngood_results, stack_data_results


    def run(self):
        if self.flag:
            start_time = time.time()
            print(f"Start stack with {self.jobs} jobs in python...")
        
        # parallel processing
        if self.jobs > 1:
            chunk_list = split(num=self.pair_num, n_jobs=self.jobs)
            results = Parallel(n_jobs=self.jobs, backend="loky")(delayed(self.process_chunk)(chunk_start, chunk_end) for chunk_start, chunk_end in chunk_list)

            ngood_all_results = []
            output_data_results = []

            for chunk_results in results:
                ngood_all_chunk, output_data_chunk = chunk_results
                ngood_all_results.extend(ngood_all_chunk)
                output_data_results.extend(output_data_chunk)

            self.ngood_all = np.array(ngood_all_results)
            self.output_data = np.array(output_data_results, dtype=self.time_dtype)
        else:
            ngood_all_results, output_data_results = self.process_chunk(0, self.pair_num)
            self.ngood_all, self.output_data = np.array(ngood_all_results), np.array(output_data_results, dtype=self.time_dtype)

        if self.flag:
            end_time = time.time()
            elapsed_time = end_time - start_time
            print(f"End stack with total time {elapsed_time}s")




import os
import h5py
import numpy as np
import matplotlib.pyplot as plt

from scipy.io import savemat
from obspy.core import UTCDateTime
from obspy.signal.filter import bandpass
from noiseflow.utils.timestamp import time_linspace, get_timestamp


class CorrData_Class(object):
    def __init__(self, corr_data, dt, 
                 corr_method, corr_pair, maxlag,smoothspect_N,
                flag, flag_gap, threads, jobs, py):
        self.corr_data = corr_data
        self.dt = dt
        self.corr_method = corr_method
        self.corr_pair = corr_pair
        self.maxlag = maxlag
        self.smoothspect_N=smoothspect_N
        self.flag=flag
        self.flag_gap=flag_gap
        self.threads=threads
        self.jobs=jobs
        self.py=py

    def save(self, save_path, format='npz', compression=False, h5_compression_format='gzip', h5_compression_opts=3):
        if format == 'npz':
            if compression:
                np.savez_compressed(save_path,
                                    corr_data=self.corr_data,
                                    dt=self.dt,
                                    corr_method = self.corr_method,
                                    corr_pair = self.corr_pair,
                                    maxlag = self.maxlag,
                                    smoothspect_N = self.smoothspect_N,
                                    flag = self.flag,
                                    flag_gap = self.flag_gap,
                                    threads = self.threads,
                                    jobs = self.jobs,
                                    py = self.py)
            else:
                np.savez(save_path,
                        corr_data=self.corr_data,
                        dt=self.dt,
                        corr_method = self.corr_method,
                        corr_pair = self.corr_pair,
                        maxlag = self.maxlag,
                        smoothspect_N = self.smoothspect_N,
                        flag = self.flag,
                        flag_gap = self.flag_gap,
                        threads = self.threads,
                        jobs = self.jobs,
                        py = self.py)
                
        elif format == 'h5':
            with h5py.File(save_path, 'w') as f:
                group = f.create_group('noiseflow_group')
                group.attrs['dt'] = self.dt
                group.attrs['corr_method'] = self.corr_method
                group.attrs['corr_pair'] = self.corr_pair
                group.attrs['maxlag'] = self.maxlag
                group.attrs['smoothspect_N'] = self.smoothspect_N
                group.attrs['flag'] = self.flag
                group.attrs['flag_gap'] = self.flag_gap
                group.attrs['threads'] = self.threads
                group.attrs['jobs'] = self.jobs
                group.attrs['py'] = self.py
                group.create_dataset('corr_pair', data=self.corr_pair)
                if compression:
                    group.create_dataset('corr_data', data=self.corr_data, compression=h5_compression_format, compression_opts=h5_compression_opts)
                else:
                    group.create_dataset('corr_data', data=self.corr_data)

        elif format == 'mat':
            savemat(save_path,
                {'corr_data': self.corr_data,
                 'dt': self.dt,
                 'corr_method': self.corr_method,
                 'corr_pair': self.corr_pair,
                 'maxlag': self.maxlag,
                 'smoothspect_N': self.smoothspect_N,
                 'flag': self.flag,
                 'flag_gap': self.flag_gap,
                 'threads': self.threads,
                 'jobs': self.jobs,
                 'py': self.py},
                do_compression=compression) 
        else:
            raise ValueError("format must be 'npz', 'h5' or 'mat'")
        

    # plot
    def plot(self, 
             pair_indx=0, 
             t_min=UTCDateTime("1970-01-01T00:00:00.0"), 
             cc_len=None, 
             cc_step=None, 
             win_start=None, 
             win_end=None, 
             lag_start=None, 
             lag_end=None, 
             amp_normalize=True, 
             amp_scale=1, 
             filter=False, 
             f1=None, 
             f2=None, 
             corners=4, 
             zerophase=True, 
             win_interval=None, 
             mode='waveform', 
             cmap='seismic',
             linewidth=0.8, 
             yticklabel_num=5, 
             figsize=(10, 6),
             save=False, 
             save_path=None, 
             dpi=100):
        
        # check pair_indx
        if pair_indx >= self.corr_data.shape[0]:
            raise ValueError("pair_indx must be <= (corr_data.shape[0]=%d)" % self.corr_data.shape[0])
        
        # init cc_len cc_step
        if cc_len == None:
            cc_len = self.corr_data.shape[2]*self.dt
        if cc_step == None:
            cc_step = 0.0

        # init timestamp
        t_max, win_interval_raw, win_vector = get_timestamp(win_num=self.corr_data.shape[1], 
                                                        cc_len=cc_len, 
                                                        cc_step=cc_step, 
                                                        t_min=t_min)

        # init win_start
        if win_start == None:
            win_start = t_min
        if win_start < t_min:
            raise ValueError("win_start must be >= (t_min=%s)" % str(t_min))

        # init win_end
        if win_end == None:
            win_end = t_max
        if win_end > t_max:
            raise ValueError("win_end must be <= (t_max=%s)" % str(t_max))

        # init win_interval
        if win_interval == None:
            win_interval = win_interval_raw      
        if win_interval < win_interval_raw:
            raise ValueError("win_interval must be >= %f" % win_interval_raw)
        if win_interval > t_max-t_min-cc_len:
            raise ValueError("win_interval must be <= %f" % (t_max-t_min-cc_len))

        # init lag_start 
        if lag_start == None:
            lag_start = -self.maxlag
        if lag_start < -self.maxlag:
            raise ValueError("lag_start must be >= (-maxlag=-%f)" % self.maxlag)
        
        # init lag_end
        if lag_end == None:
            lag_end = self.maxlag
        if lag_end > self.maxlag:
            raise ValueError("lag_end must be <= (maxlag=%f)" % self.maxlag)
        
        # lag vector
        tt = np.arange(lag_start, lag_end+self.dt, self.dt)
        tp_start = round((lag_start+self.maxlag)/self.dt)
        tp_end = tp_start+len(tt)  # make sure same length, so do not use round((lag_end+self.maxlag+self.dt)/self.dt)

        # select and filter data according to win_start, win_end
        win_indx = np.where((win_vector >= win_start) & (win_vector <= win_end))[0]
        if filter:
            data=np.empty((len(win_indx), len(tt)))
            for i in range(0, len(win_indx)):
                data[i,:] = bandpass(self.corr_data[pair_indx, win_indx[i], tp_start:tp_end], f1, f2, int(1/self.dt), corners=corners, zerophase=zerophase)
        else:
            data = self.corr_data[pair_indx, win_indx, tp_start:tp_end]

        # plot
        if mode == 'waveform':
            # init fig
            fig, ax = plt.subplots(figsize=figsize)

            # normalize
            if amp_normalize:
                scale = np.max(np.abs(data), axis=1)[:,None]
                if np.min(scale) != 0:
                    data_plot = amp_scale*data/scale
                else:
                    raise ValueError("data is all zeros in amp_normalize=True")
            else:
                scale = np.max(np.abs(data))
                if scale != 0:
                    data_plot = amp_scale*data/scale
                else:
                    raise ValueError("data is all zeros in amp_normalize=False")

            # get win_interval index
            ydist_n=[]
            win_interval_n = round(win_interval/win_interval_raw)
            for i in range(0, len(win_indx), win_interval_n):
                ydist_n.append(i)
            ydist_n = np.int64(np.array(ydist_n))

            # plot wavefrom
            for i in range(0, len(ydist_n)):
                ax.plot(tt, data_plot[ydist_n[i]]+ydist_n[i], 'k', linewidth=linewidth)

            # set x axis
            ax.set_title("CorrData: pair_indx=%d, filter=[%.2f, %.2f] hz" % (pair_indx, f1, f2))
            ax.set_xlim(tt[0], tt[-1])
            ax.set_xlabel('Time(s)')
            
            # set y axis
            yy_tick = np.linspace(0, len(win_indx)-1, num=yticklabel_num)
            yy_label_UTC = time_linspace(win_start, win_end, num=yticklabel_num)
            yy_label = np.array([i.strftime('%Y-%m-%dT%H:%M:%S') for i in yy_label_UTC])
            ax.set_yticks(yy_tick)
            ax.set_yticklabels(yy_label, rotation=45)

            # save or show
            if save:
                if save_path is None:
                    raise ValueError("save_path must be specified")
                fig.savefig(os.path.join(save_path), dpi=dpi, format='pdf', bbox_inches='tight')
                plt.close(fig)
            else:
                plt.show()

        elif mode == 'mat':
            # init fig
            fig, ax = plt.subplots(figsize=figsize)

            # average all data
            cc_mean = np.mean(data, axis=0)/np.max(np.mean(data, axis=0))

            # average win_interval data
            ntrace = int((win_end-win_start)/win_interval)
            ndata  = np.zeros(shape=(ntrace,len(tt)))
            for i in range(0,ntrace):
                tindx = np.where(((win_vector-win_start)>=i*win_interval) & ((win_vector-win_start)<(i+1)*win_interval))[0]
                ndata[i] = np.mean(data[tindx],axis=0)

            # normalize waveforms
            if amp_normalize:
                scale = np.max(np.abs(ndata), axis=1)[:,None]
                if np.min(scale) != 0:
                    ndata_plot = ndata/scale
                else:
                    raise ValueError("data is all zeros in amp_normalize=True")
            else:
                scale = np.max(np.abs(ndata))
                if scale != 0:
                    ndata_plot = ndata/scale
                else:
                    raise ValueError("data is all zeros in amp_normalize=False")
                
            # plot mat
            cax=ax.matshow(ndata_plot, extent=[lag_start, lag_end, ntrace, 0], aspect='auto', cmap=cmap)
            ax.plot(tt, ntrace/10*cc_mean+ntrace/2, 'k', linewidth=linewidth)
            
            # set x axis
            ax.set_title("CorrData: pair_indx=%d, filter=[%.2f, %.2f] hz" % (pair_indx, f1, f2))
            ax.set_xlim(tt[0], tt[-1])
            ax.set_xlabel('Time(s)')
            ax.xaxis.set_ticks_position('bottom')
            
            # set y axis
            ax.invert_yaxis()
            ax.set_ylim(0, ntrace)
            yy_tick=np.linspace(0, ntrace, num=yticklabel_num)
            yy_label_UTC=time_linspace(win_start, win_end, num=yticklabel_num)
            yy_label = np.array([i.strftime('%Y-%m-%dT%H:%M:%S') for i in yy_label_UTC])
            ax.set_yticks(yy_tick)
            ax.set_yticklabels(yy_label, rotation=45)

            # save or show
            if save:
                if save_path is None:
                    raise ValueError("save_path must be specified")
                fig.savefig(os.path.join(save_path), dpi=dpi, format='pdf', bbox_inches = 'tight') 
                plt.close(fig)
            else:
                plt.show()

        else:
            raise ValueError("mode must be 'waveform' or 'mat'")
        



import warnings

from scipy.signal import iirfilter
from scipy.signal import sosfilt
from scipy.signal import zpk2sos



def bandpass_py(data, freqmin, freqmax, df, corners=4, zerophase=True, flag=False, flag_gap=None):
    if data.ndim == 1:
        data = data.reshape(1, -1)

    for i in range(0, data.shape[0]):
        data[i,:] = bandpass(data[i,:], freqmin, freqmax, df, corners, zerophase)


def bandstop_py(data, freqmin, freqmax, df, corners=4, zerophase=True, flag=False, flag_gap=None):
    if data.ndim == 1:
        data = data.reshape(1, -1)

    for i in range(0, data.shape[0]):
        data[i,:] = bandstop(data[i,:], freqmin, freqmax, df, corners, zerophase)


def lowpass_py(data, freq, df, corners=4, zerophase=True, flag=False, flag_gap=None):
    if data.ndim == 1:
        data = data.reshape(1, -1)

    for i in range(0, data.shape[0]):
        data[i,:] = lowpass(data[i,:], freq, df, corners, zerophase)


def highpass_py(data, freq, df, corners=4, zerophase=True, flag=False, flag_gap=None):
    if data.ndim == 1:
        data = data.reshape(1, -1)
        
    for i in range(0, data.shape[0]):
        data[i,:] = highpass(data[i,:], freq, df, corners, zerophase)



def bandpass(data, freqmin, freqmax, df, corners=4, zerophase=False):
    """
    Butterworth-Bandpass Filter.
    Filter data from ``freqmin`` to ``freqmax`` using ``corners``
    corners.
    The filter uses :func:`scipy.signal.iirfilter` (for design)
    and :func:`scipy.signal.sosfilt` (for applying the filter).
    :type data: numpy.ndarray
    :param data: Data to filter.
    :param freqmin: Pass band low corner frequency.
    :param freqmax: Pass band high corner frequency.
    :param df: Sampling rate in Hz.
    :param corners: Filter corners / order.
    :param zerophase: If True, apply filter once forwards and once backwards.
        This results in twice the filter order but zero phase shift in
        the resulting filtered trace.
    :return: Filtered data.
    """
    fe = 0.5 * df
    low = freqmin / fe
    high = freqmax / fe
    # raise for some bad scenarios
    if high - 1.0 > -1e-6:
        msg = ("Selected high corner frequency ({}) of bandpass is at or "
               "above Nyquist ({}). Applying a high-pass instead.").format(
            freqmax, fe)
        warnings.warn(msg)
        return highpass(data, freq=freqmin, df=df, corners=corners,
                        zerophase=zerophase)
    if low > 1:
        msg = "Selected low corner frequency is above Nyquist."
        raise ValueError(msg)
    z, p, k = iirfilter(corners, [low, high], btype='band',
                        ftype='butter', output='zpk')
    sos = zpk2sos(z, p, k)
    if zerophase:
        firstpass = sosfilt(sos, data)
        return sosfilt(sos, firstpass[::-1])[::-1]
    else:
        return sosfilt(sos, data)


def bandstop(data, freqmin, freqmax, df, corners=4, zerophase=False):
    """
    Butterworth-Bandstop Filter.
    Filter data removing data between frequencies ``freqmin`` and ``freqmax``
    using ``corners`` corners.
    The filter uses :func:`scipy.signal.iirfilter` (for design)
    and :func:`scipy.signal.sosfilt` (for applying the filter).
    :type data: numpy.ndarray
    :param data: Data to filter.
    :param freqmin: Stop band low corner frequency.
    :param freqmax: Stop band high corner frequency.
    :param df: Sampling rate in Hz.
    :param corners: Filter corners / order.
    :param zerophase: If True, apply filter once forwards and once backwards.
        This results in twice the number of corners but zero phase shift in
        the resulting filtered trace.
    :return: Filtered data.
    """
    fe = 0.5 * df
    low = freqmin / fe
    high = freqmax / fe
    # raise for some bad scenarios
    if high > 1:
        high = 1.0
        msg = "Selected high corner frequency is above Nyquist. " + \
              "Setting Nyquist as high corner."
        warnings.warn(msg)
    if low > 1:
        msg = "Selected low corner frequency is above Nyquist."
        raise ValueError(msg)
    z, p, k = iirfilter(corners, [low, high],
                        btype='bandstop', ftype='butter', output='zpk')
    sos = zpk2sos(z, p, k)
    if zerophase:
        firstpass = sosfilt(sos, data)
        return sosfilt(sos, firstpass[::-1])[::-1]
    else:
        return sosfilt(sos, data)


def lowpass(data, freq, df, corners=4, zerophase=False):
    """
    Butterworth-Lowpass Filter.
    Filter data removing data over certain frequency ``freq`` using ``corners``
    corners.
    The filter uses :func:`scipy.signal.iirfilter` (for design)
    and :func:`scipy.signal.sosfilt` (for applying the filter).
    :type data: numpy.ndarray
    :param data: Data to filter.
    :param freq: Filter corner frequency.
    :param df: Sampling rate in Hz.
    :param corners: Filter corners / order.
    :param zerophase: If True, apply filter once forwards and once backwards.
        This results in twice the number of corners but zero phase shift in
        the resulting filtered trace.
    :return: Filtered data.
    """
    fe = 0.5 * df
    f = freq / fe
    # raise for some bad scenarios
    if f > 1:
        f = 1.0
        msg = "Selected corner frequency is above Nyquist. " + \
              "Setting Nyquist as high corner."
        warnings.warn(msg)
    z, p, k = iirfilter(corners, f, btype='lowpass', ftype='butter',
                        output='zpk')
    sos = zpk2sos(z, p, k)
    if zerophase:
        firstpass = sosfilt(sos, data)
        return sosfilt(sos, firstpass[::-1])[::-1]
    else:
        return sosfilt(sos, data)


def highpass(data, freq, df, corners=4, zerophase=False):
    """
    Butterworth-Highpass Filter.
    Filter data removing data below certain frequency ``freq`` using
    ``corners`` corners.
    The filter uses :func:`scipy.signal.iirfilter` (for design)
    and :func:`scipy.signal.sosfilt` (for applying the filter).
    :type data: numpy.ndarray
    :param data: Data to filter.
    :param freq: Filter corner frequency.
    :param df: Sampling rate in Hz.
    :param corners: Filter corners / order.
    :param zerophase: If True, apply filter once forwards and once backwards.
        This results in twice the number of corners but zero phase shift in
        the resulting filtered trace.
    :return: Filtered data.
    """
    fe = 0.5 * df
    f = freq / fe
    # raise for some bad scenarios
    if f > 1:
        msg = "Selected corner frequency is above Nyquist."
        raise ValueError(msg)
    z, p, k = iirfilter(corners, f, btype='highpass', ftype='butter',
                        output='zpk')
    sos = zpk2sos(z, p, k)
    if zerophase:
        firstpass = sosfilt(sos, data)
        return sosfilt(sos, firstpass[::-1])[::-1]
    else:
        return sosfilt(sos, data)









import scipy
import numpy as np


def decimate_py(data, df, df_new, flag=False, flag_gap=None):
    if data.ndim == 1:
        data = data.reshape(1, -1)

    test = decimate(data[0,:], df, df_new)
    redata = np.empty((data.shape[0], test.shape[0]))

    for i in range(0, data.shape[0]):
        redata[i,:] = decimate(data[i,:], df, df_new)

    return redata


def decimate(data, df, df_new):
    factor = int(df / df_new)
    if factor > 13:
        msg = (
                "IRR filter is unstable for decimation factors above"
                " 13. Call decimate multiple times."
            )
        raise ValueError(msg)
    
    redata = scipy.signal.decimate(data, factor, ftype='iir', axis=-1, zero_phase=True)

    return redata







import scipy
import warnings
import numpy as np


def taper_py(data, max_percentage=0.05, type='hann', side='both', flag=False, flag_gap=None):
    if data.ndim == 1:
        data = data.reshape(1, -1)

    for i in range(0, data.shape[0]):
        taper(data[i,:], max_percentage, type, side)



def taper(data, max_percentage=0.05, type='hann', side='both'):

    side_valid = ['both', 'left', 'right']
    npts = len(data)
    if side not in side_valid:
        raise ValueError("'side' has to be one of: %s" % side_valid)

    # max_half_lenghts_1
    max_half_lenghts_1 = None
    if max_percentage is not None:
        max_half_lenghts_1 = int(max_percentage * npts)

    if 2 * max_half_lenghts_1 > npts:
        msg = "The requested taper is longer than the data. " \
                "The taper will be shortened to data length."
        warnings.warn(msg)

    # max_half_lenghts_2
    max_half_lenghts_2 = int(npts / 2)
    if max_half_lenghts_1 is None:
        wlen = max_half_lenghts_2
    else:
        wlen = min(max_half_lenghts_1, max_half_lenghts_2)

    if type == "hann":
        taper_sides = scipy.signal.windows.hann(2*wlen+1)
    elif type == "hamming":
        taper_sides = scipy.signal.windows.hamming(2*wlen+1)
    elif type == "bartlett":
        taper_sides = scipy.signal.windows.bartlett(2*wlen+1)
    elif type == "blackman":
        taper_sides = scipy.signal.windows.blackman(2*wlen+1)
    elif type == "flattop":
        taper_sides = scipy.signal.windows.flattop(2*wlen+1)
    elif type == "parzen":
        taper_sides = scipy.signal.windows.parzen(2*wlen+1)
    elif type == "bohman":
        taper_sides = scipy.signal.windows.bohman(2*wlen+1)
    elif type == "blackmanharris":
        taper_sides = scipy.signal.windows.blackmanharris(2*wlen+1)
    elif type == "nuttall":
        taper_sides = scipy.signal.windows.nuttall(2*wlen+1)
    elif type == "barthann":
        taper_sides = scipy.signal.windows.barthann(2*wlen+1)
    elif type == "kaiser":
        taper_sides = scipy.signal.windows.kaiser(2*wlen+1, beta=14)
    elif type == "gaussian":
        taper_sides = scipy.signal.windows.gaussian(2*wlen+1, std=14)
    elif type == "general_gaussian":
        taper_sides = scipy.signal.windows.general_gaussian(2*wlen+1, p=1, sig=14)
    else:
        raise ValueError("Unknown taper type: %s" % type)

    if side == 'left':
        taper = np.hstack((taper_sides[:wlen], np.ones(npts - wlen)))
    elif side == 'right':
        taper = np.hstack((np.ones(npts - wlen),
                            taper_sides[len(taper_sides) - wlen:]))
    else:
        taper = np.hstack((taper_sides[:wlen], np.ones(npts - 2 * wlen),
                            taper_sides[len(taper_sides) - wlen:]))

    # Convert data if it's not a floating point type.
    if not np.issubdtype(data.dtype, np.floating):
        data = np.require(data, dtype=np.float64)

    data *= taper







import numpy as np


def split(num, n_jobs):
    chunk_size = num // n_jobs
    results = []
    for i in range(0, n_jobs):
        chunk_start = i*chunk_size
        if i == n_jobs - 1:
            chunk_end = num
        else:
            chunk_end = (i+1)*chunk_size
        results.append([chunk_start, chunk_end])

    return results


def slice_window(npts, segment_points, step_points):
    if segment_points < npts:
        slide_points = segment_points - step_points
        win_num = 0
        for i in range(0, int(npts/slide_points)):
            if (i * slide_points + segment_points) <= npts:
                win_num += 1
            else:
                break
        win_info = np.empty((win_num, 2), dtype=int)
        for i in range(win_num):
            win_info[i, 0] = i * slide_points
            win_info[i, 1] = i * slide_points + segment_points
    elif segment_points == npts:
        win_num = 1
        win_info = np.array([[0, npts]], dtype=int)
    else:
        raise ValueError("error: segment-points length is larger than npts when slicing windows!")
    
    return win_info


def moving_ave(A, N):
    temp = np.zeros(A.shape[0] + 2*N)

    temp_len = temp.shape[0]
    temp[N:temp_len-N] = A
    temp[0:N] = temp[N]
    temp[temp_len-N:temp_len] = temp[temp_len-N-1]

    nn = np.ones(N) / N
    b1 = np.convolve(temp, nn, mode='full')

    n1 = N + (N-1) // 2
    n2 = N + (N-1 - (N-1) // 2)
    B1 = b1[n1 : b1.shape[0] - n2]

    return B1


def whiten(rfft_data, dt, freq_norm_method, freqmin, freqmax, smoothspect_N, whiten_npad):
    _i = 0.0 + 1.0j
    _i0 = 0.0 + 0.0j
    _pi = np.pi

    win_num = rfft_data.shape[0]
    fft_npts = rfft_data.shape[1]

    freq_array = np.arange(fft_npts) / dt / 2.0 / (fft_npts-1) # note: must divide by 2.0, because rfft_npts is not npts.
    J = np.where((freq_array>=freqmin) & (freq_array<=freqmax))[0]

    low = J[0] - whiten_npad
    if (low <= 0):
        low = 0
    left = J[0]
    right = J[-1]
    high = J[-1] + whiten_npad
    if (high > fft_npts):
        high = fft_npts

    rfft_norm_data = _i0*np.ones((win_num, fft_npts))

    for i in range(0, win_num):
        #  left zero cut-off
        # rfft_norm_data[i, 0:low] = _i0*rfft_data[i, 0:low]

        # // left tapering
        smo1 = np.power(np.cos( np.linspace(_pi/2, _pi, left-low)), 2)
        exp1 = np.exp(_i*np.angle(rfft_data[i, low:left]))
        rfft_norm_data[i, low:left] = smo1*exp1
 
        # // pass band
        if freq_norm_method == "whiten":
            smo2 = np.ones((right-left))
            exp2 = np.exp(_i*np.angle(rfft_data[i, left: right]))
            rfft_norm_data[ i, left:right] = smo2*exp2
        elif freq_norm_method == "smooth_whiten":
            data = rfft_data[i, left:right]
            rfft_norm_data[ i, left:right] = data / moving_ave(np.abs(data), smoothspect_N)
    
        # // right tapering
        smo3 = np.power(np.cos( np.linspace(0, _pi/2, high-right)), 2)
        exp3 = np.exp(_i*np.angle(rfft_data[ i, right: high]))
        rfft_norm_data[i, right: high] = smo3*exp3

        # // right zero cut-off
        # rfft_norm_data[i, high:fft_npts] = _i0*rfft_data[i, high:fft_npts]
    
    return rfft_norm_data



import numpy as np

from scipy.signal import detrend as scipy_detrend
from scipy.interpolate import LSQUnivariateSpline


def detrend_py(data, type='linear', flag=False, flag_gap=None):
    if data.ndim == 1:
        data = data.reshape(1, -1)

    for i in range(0, data.shape[0]):
        detrend(data[i,:], type)


def spline(data, order=2, dspline=1000):
    if not np.issubdtype(data.dtype, np.floating):
        data = np.require(data, dtype=np.float64)

    x = np.arange(len(data))
    splknots = np.arange(dspline/2.0, len(data) - dspline/2.0 + 2, dspline)
    spl = LSQUnivariateSpline(x=x, y=data, t=splknots, k=order)
    fit = spl(x)
    data -= fit

    return data


def detrend(data, type='linear', order=2, dspline=1000):
    """
    Detrend data.
    :param data: The data to detrend.
    :type data: :class:`numpy.ndarray`
    :param type: The type of detrending to perform. Can be one of ``'demean'``,
        ``'linear'``, or ``'spline'``.
    :type type: str
    :param overwrite_data: If True the data will be detrended in place.
    :type overwrite_data: bool
    :param dtype: The data type to use for the detrended data. Defaults to
        ``np.float32``.
    :type dtype: :class:`numpy.dtype`
    :param order: The order of the spline to fit. Defaults to 2.
    :type order: int
    :param dspline: The distance between spline knots. Defaults to 1000.
    :type dspline: int

    :returns: The detrended data.
    :rtype: :class:`numpy.ndarray`

    .. rubric:: Example

    >>> import numpy as np
    >>> from noiseflow.signal import detrend
    >>> data = np.array([1.1,2.7,3,3,3,6,7,8,9,10], overwrite_data=True)
    >>> detrend(data, type='spline')

    """

    if type == 'demean':
        data -= np.mean(data)
    elif type == 'linear':
        scipy_detrend(data, type='linear', overwrite_data=True)
    elif type == 'spline':
        spline(data, order=order, dspline=dspline)
    else:
        raise ValueError("type must be 'demean', 'linear', or 'spline'")









import os
import h5py
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors

from scipy.io import savemat
from obspy.core import UTCDateTime
from obspy.signal.filter import bandpass
from noiseflow.utils.timestamp import time_linspace


class StackData_Class(object):
    def __init__(self, stack_data, stack_ngood, dt,
                 stack_method, par, stack_all, stack_len, stack_step, pick, median_high, median_low, flag, flag_gap, threads, jobs, py):
        self.stack_data = stack_data
        self.stack_ngood = stack_ngood
        self.dt = dt
        self.stack_method = stack_method
        self.par = par
        self.stack_all = stack_all
        self.stack_len = stack_len
        self.stack_step = stack_step
        self.pick = pick
        self.median_high = median_high
        self.median_low = median_low
        self.flag=flag
        self.flag_gap=flag_gap
        self.threads=threads
        self.jobs=jobs
        self.py=py

    # save data
    def save(self, save_path, format='npz', compression=False, h5_compression_format='gzip', h5_compression_opts=3):
        self.maxlag = self.dt * (self.stack_data.shape[2]-1)/2

        if format == 'npz':
            if compression:
                np.savez_compressed(save_path,
                                    stack_data = self.stack_data,
                                    stack_ngood = self.stack_ngood,
                                    dt = self.dt,
                                    stack_method = self.stack_method,
                                    par = self.par,
                                    stack_all = self.stack_all,
                                    stack_len = self.stack_len,
                                    stack_step = self.stack_step,
                                    pick = self.pick,
                                    median_high = self.median_high,
                                    median_low = self.median_low,
                                    flag = self.flag,
                                    flag_gap = self.flag_gap,
                                    threads = self.threads,
                                    jobs = self.jobs,
                                    py = self.py,
                                    maxlag = self.maxlag)
            else:
                np.savez(save_path,
                        stack_data = self.stack_data,
                        stack_ngood = self.stack_ngood,
                        dt = self.dt,
                        stack_method = self.stack_method,
                        par = self.par,
                        stack_all = self.stack_all,
                        stack_len = self.stack_len,
                        stack_step = self.stack_step,
                        pick = self.pick,
                        median_high = self.median_high,
                        median_low = self.median_low,
                        flag = self.flag,
                        flag_gap = self.flag_gap,
                        threads = self.threads,
                        jobs = self.jobs,
                        py = self.py,
                        maxlag = self.maxlag)
                
        elif format == 'h5':
            par_copy = self.par.copy()
            with h5py.File(save_path, 'w') as f:
                group = f.create_group('noiseflow_group')
                group.attrs['dt'] = self.dt
                group.attrs['stack_method'] = self.stack_method
                group.attrs['stack_all'] = self.stack_all
                group.attrs['stack_len'] = self.stack_len
                group.attrs['stack_step'] = self.stack_step
                group.attrs['pick'] = self.pick
                group.attrs['median_high'] = self.median_high
                group.attrs['median_low'] = self.median_low
                group.attrs['flag'] = self.flag
                group.attrs['flag_gap'] = self.flag_gap
                group.attrs['threads'] = self.threads
                group.attrs['jobs'] = self.jobs
                group.attrs['py'] = self.py
                group.attrs['maxlag'] = self.maxlag

                for key, value in par_copy.items():
                    if value is None:
                        value = float('nan')
                    group.attrs[key] = value

                group.create_dataset('stack_ngood', data=self.stack_ngood)
                if compression:
                    group.create_dataset('stack_data', data=self.stack_data, compression=h5_compression_format, compression_opts=h5_compression_opts)
                else:
                    group.create_dataset('stack_data', data=self.stack_data)

        elif format == 'mat':
            par_copy = self.par.copy()
            for key, value in par_copy.items():
                if value is None:
                    par_copy[key] = np.nan

            savemat(save_path,
                {'stack_data': self.stack_data,
                'stack_ngood': self.stack_ngood,
                'dt': self.dt,
                'stack_method': self.stack_method,
                'par': par_copy,
                'stack_all': self.stack_all,
                'stack_len': self.stack_len,
                'stack_step': self.stack_step,
                'pick': self.pick,
                'median_high': self.median_high,
                'median_low': self.median_low,
                'flag': self.flag,
                'flag_gap': self.flag_gap,
                'threads': self.threads,
                'jobs': self.jobs,
                'py': self.py,
                'maxlag': self.maxlag},
                do_compression=compression)
        else:
            raise ValueError("format must be 'npz', 'h5' or 'mat'")
        

    # plot
    def plot(self, 
             pair_indx=0, 
             t_min=UTCDateTime("1970-01-01T00:00:00.0"), 
             stack_len=1, 
             stack_step=0, 
             cc_len=None, 
             cc_step=None, 
             win_start=None, 
             win_end=None, 
             lag_start=None, 
             lag_end=None, 
             amp_normalize=True, 
             amp_scale=1, 
             filter=False, 
             f1=None, 
             f2=None, 
             corners=4, 
             zerophase=True, 
             win_interval=None, 
             mode='waveform', 
             cmap='seismic', 
             linewidth=0.8,
             yticklabel_num=5, 
             figsize=(10, 6),
             ngood_label=False, 
             save=False, 
             save_path=None, 
             dpi=300): 
        
        self.maxlag = self.dt * (self.stack_data.shape[2]-1)/2

        # check pair_indx
        if pair_indx >= self.stack_data.shape[0]:
            raise ValueError("pair_indx must be smaller than stack_data.shape[0]=%d" % self.stack_data.shape[0])

        # init cc_len cc_step
        if cc_len == None:
            cc_len = self.stack_data.shape[2]*self.dt
        if cc_step == None:
            cc_step = 0.0

        # define time vector
        stack_win_num = self.stack_data.shape[1]
        corr_win_num = (stack_len-stack_step)*(stack_win_num-1) + stack_len
        t_max = t_min + (corr_win_num-1)*(cc_len-cc_step) + cc_len
        stack_interval = (stack_len-stack_step)*(cc_len-cc_step)
        win_vector = np.array([t_min + i*stack_interval for i in range(0, stack_win_num)])

        # init win_start
        if win_start == None:
            win_start = t_min
        if win_start < t_min:
            raise ValueError("win_start must be larger than t_min=%s" % str(t_min))

        # init win_end
        if win_end == None:
            win_end = t_max
        if win_end > t_max:
            raise ValueError("win_end must be smaller than t_max=%s" % str(t_max))

        # init lag_start 
        if lag_start == None:
            lag_start = -self.maxlag
        if lag_start < -self.maxlag:
            raise ValueError("lag_start must be larger than -maxlag=-%f" % self.maxlag)
        
        # init lag_end
        if lag_end == None:
            lag_end = self.maxlag
        if lag_end > self.maxlag:
            raise ValueError("lag_end must be smaller than maxlag=%f" % self.maxlag)
        
        # lag vector
        tt = np.arange(lag_start, lag_end+self.dt, self.dt)
        tp_start = round((lag_start+self.maxlag)/self.dt)
        tp_end = tp_start+len(tt)  # make sure same length, so do not use round((lag_end+self.maxlag+self.dt)/self.dt)

        # check win_interval
        if win_interval == None:
            win_interval = (stack_len-stack_step)*(cc_len-cc_step)
        if win_interval < (stack_len-stack_step)*(cc_len-cc_step):
            raise ValueError("win_interval must be larger than (stack_len-stack_step)*(cc_len-cc_step)=%f" % (stack_len-stack_step)*(cc_len-cc_step))
        if win_interval > t_max-stack_len*(cc_len-cc_step)-t_min:
            raise ValueError("win_interval must be smaller than t_max-stack_len*(cc_len-cc_step)-t_min=%f" % (t_max-stack_len*(cc_len-cc_step)-t_min))
        
        # select win_indx data and filter --> data, ngood
        win_indx = np.where((win_vector >= win_start) & (win_vector <= win_end))[0]
        ngood = self.stack_ngood[pair_indx, win_indx]
        if filter:
            data=np.empty((len(win_indx), len(tt)))
            for i in range(0, len(win_indx)):
                data[i,:] = bandpass(self.stack_data[pair_indx, win_indx[i], tp_start:tp_end], f1, f2, int(1/self.dt), corners=corners, zerophase=zerophase)
        else:
            data = self.stack_data[pair_indx, win_indx, tp_start:tp_end]

        # plot
        if mode == 'waveform':
            # init figure
            fig, ax = plt.subplots(1, 2, gridspec_kw=dict(width_ratios=[5, 1]), sharey="row", figsize=figsize)
            fig.subplots_adjust(wspace=0.01)

            # normalize
            if amp_normalize:
                scale = np.max(np.abs(data), axis=1)[:,None]
                if np.min(scale) != 0:
                    data_plot = amp_scale*data/scale
                else:
                    raise ValueError("data is all zeros in amp_normalize=True")
            else:
                scale = np.max(np.abs(data))
                if scale != 0:
                    data_plot = amp_scale*data/scale
                else:
                    raise ValueError("data is all zeros in amp_normalize=False")

            # select win_interval data
            ydist_n=[]
            ngood_n=[]
            win_interval_n = round(win_interval/((stack_len-stack_step)*(cc_len-cc_step)))
            for i in range(0, len(win_indx), win_interval_n):
                ydist_n.append(i)
                ngood_n.append(ngood[i])
            ydist_n = np.array(ydist_n)
            ngood_n = np.array(ngood_n)

            # plot wavefrom
            for i in range(0, len(ydist_n)):
                ax[0].plot(tt, data_plot[int(ydist_n[i])]+ydist_n[i], 'k', linewidth=linewidth)
            
            # set x axis
            ax[0].set_title("StackData: pair_indx=%d, filter=[%.2f, %.2f] hz" % (pair_indx, f1, f2))
            ax[0].set_xlim(tt[0], tt[-1])
            ax[0].set_xlabel('Time(s)')
            
            # set y axis
            yy_tick = np.linspace(0, len(win_indx)-1, num=yticklabel_num)
            yy_label_UTC = time_linspace(win_start, win_end, num=yticklabel_num)
            yy_label = np.array([i.strftime('%Y-%m-%dT%H:%M:%S') for i in yy_label_UTC])
            ax[0].set_yticks(yy_tick)
            ax[0].set_yticklabels(yy_label, rotation=45)

            # plot ngood 
            (markers, stemlines, baseline) = ax[1].stem(ydist_n, ngood_n, orientation='horizontal')
            plt.setp(markers, marker='D', markeredgecolor="orange", markeredgewidth=1.5)
            plt.setp(stemlines, linestyle="-", color="olive", linewidth=linewidth)
            if ngood_label:
                for i in range(0, len(ngood_n)):
                    ax[1].text(ngood_n[i]+np.max(ngood_n)/10, ydist_n[i], int(ngood_n[i]), va='center', ha='left')

            ax[1].set_xlim(0, np.max(ngood_n)*1.2)
            ax[1].set_xlabel('Ngood')
            ax[1].get_yaxis().set_visible(False)
            ax[1].spines['top'].set_visible(False)
            ax[1].spines['right'].set_visible(False)
            ax[1].spines['bottom'].set_visible(True)
            ax[1].spines['left'].set_visible(True)

            # save or show
            if save:
                if save_path is None:
                    raise ValueError("save_path must be specified")
                fig.savefig(os.path.join(save_path), dpi=dpi, format='pdf', bbox_inches='tight')
                plt.close(fig)
            else:
                plt.show()

        elif mode == 'mat':
            # init figure
            fig, ax = plt.subplots(1, 2, gridspec_kw=dict(width_ratios=[5, 1]), sharey="row", figsize=figsize)
            fig.subplots_adjust(wspace=0.01)

            # average waveforms
            cc_mean = np.mean(data, axis=0)/np.max(np.mean(data, axis=0))
            ntrace = int((win_end-win_start)/win_interval)
            ndata  = np.zeros(shape=(ntrace,len(tt)))
            for i in range(0,ntrace):
                tindx = np.where(((win_vector-win_start)>=i*win_interval) & ((win_vector-win_start)<(i+1)*win_interval))[0]
                ndata[i] = np.mean(data[tindx],axis=0)

            # normalize waveforms
            if amp_normalize:
                scale = np.max(np.abs(ndata), axis=1)[:,None]
                if np.min(scale) != 0:
                    ndata_plot = ndata/scale
                else:
                    raise ValueError("data is all zeros in amp_normalize=True")
            else:
                scale = np.max(np.abs(ndata))
                if scale != 0:
                    ndata_plot = ndata/scale
                else:
                    raise ValueError("data is all zeros in amp_normalize=False")
                
            ### plot mat
            cax=ax[0].matshow(ndata_plot, extent=[lag_start, lag_end, ntrace, 0], aspect='auto', cmap=cmap)
            ax[0].plot(tt,ntrace/10*cc_mean+ntrace/2,'k',linewidth=linewidth)
            
            # set x axis
            ax[0].set_title("StackData: pair_indx=%d, filter=[%.2f, %.2f] hz" % (pair_indx, f1, f2))
            ax[0].set_xlim(tt[0], tt[-1])
            ax[0].set_xlabel('Time(s)')
            ax[0].xaxis.set_ticks_position('bottom')

            # set y axis
            ax[0].invert_yaxis()
            ax[0].set_ylim(0, ntrace)
            yy_tick = np.linspace(0, ntrace, num=yticklabel_num)
            yy_label_UTC = time_linspace(win_start, win_end, num=yticklabel_num)
            yy_label = np.array([i.strftime('%Y-%m-%dT%H:%M:%S') for i in yy_label_UTC])
            ax[0].set_yticks(yy_tick)
            ax[0].set_yticklabels(yy_label, rotation=45)

            ### plot ngood
            ngood = self.stack_ngood[pair_indx, win_indx]
            yy = np.linspace(0, ntrace, len(ngood))
            (markers, stemlines, baseline) = ax[1].stem(yy, ngood, orientation='horizontal')
            plt.setp(markers, marker='D', markeredgecolor="orange", markeredgewidth=1.5)
            plt.setp(stemlines, linestyle="-", color="olive", linewidth=linewidth)
            if ngood_label:
                for i in range(0, len(ngood)):
                    ax[1].text(ngood[i]+np.max(ngood)/10, yy[i], int(ngood[i]), va='center', ha='left')

            ax[1].set_xlim(0, np.max(ngood)*1.2)
            ax[1].set_xlabel('Ngood')
            ax[1].get_yaxis().set_visible(False)
            ax[1].spines['top'].set_visible(False)
            ax[1].spines['right'].set_visible(False)
            ax[1].spines['bottom'].set_visible(True)
            ax[1].spines['left'].set_visible(True)
            
            # save
            if save:
                if save_path is None:
                    raise ValueError("save_path must be specified")
                fig.savefig(os.path.join(save_path), dpi=dpi, format='pdf', bbox_inches = 'tight') 
                plt.close(fig)
            else:
                plt.show()

        else:
            raise ValueError("mode must be 'waveform' or 'mat'")
        


    def plot_moveout(self, 
                corr_pair, 
                pair_dist, 
                source_indx=None, 
                receiver_indx=None, 
                dist_start=None, 
                dist_end=None, 
                amp_scale=1, 
                amp_normalize=True,
                win_num = 0, 
                lag_start=None, 
                lag_end=None, 
                filter=False, 
                f1=None, 
                f2=None, 
                corners=4, 
                zerophase=True, 
                dist_interval=None, 
                mode='waveform', 
                cmap='seismic',
                linewidth=0.8, 
                yticklabel_num=10, 
                figsize=(10, 6),
                dist_unit="m", 
                velocity=[], 
                save=False, 
                save_path=None, 
                dpi=100): 
        
        self.maxlag = self.dt * (self.stack_data.shape[2]-1)/2

        # check source_indx and receiver_indx
        if source_indx != None and receiver_indx != None:
            raise ValueError("source_indx and receiver_indx cannot be specified at the same time")
        elif source_indx != None:
            if source_indx not in corr_pair[:,0]:
                raise ValueError("source_indx must be in corr_pair[:,0]")
        elif receiver_indx != None:
            if receiver_indx not in corr_pair[:,1]:
                raise ValueError("receiver_indx must be in corr_pair[:,1]")
        else:
            raise ValueError("source_indx or receiver_indx must be specified")
        
        # check corr_pair pair_dist
        if corr_pair.shape[0] != len(pair_dist):
            raise ValueError("the length of corr_pair is not consistent with pair_dist")
        
        # check demision of pair_dist
        if pair_dist.ndim == 2:
            pair_dist = pair_dist.reshape(-1)

        # check dist_unit
        if dist_unit not in ['m', 'km', 'degree']:
            raise ValueError("dist_unit must be 'm', 'km', or 'degree'")
        
        # init dist_start
        if dist_start == None:
            dist_start = np.min(pair_dist)
        if dist_start < np.min(pair_dist):
            raise ValueError("dist_start must be >= (np.min(pair_dist)=%f)" % np.min(pair_dist))
        
        # init dist_end
        if dist_end == None:
            dist_end = np.max(pair_dist)
        if dist_end > np.max(pair_dist):
            raise ValueError("dist_end must be <= (np.max(pair_dist)=%f)" % np.max(pair_dist))

        # init lag_start 
        if lag_start == None:
            lag_start = -self.maxlag
        if lag_start < -self.maxlag:
            raise ValueError("lag_start must be >= (-maxlag=-%f)" % self.maxlag)
        
        # init lag_end
        if lag_end == None:
            lag_end = self.maxlag
        if lag_end > self.maxlag:
            raise ValueError("lag_end must be <= (maxlag=%f)" % self.maxlag)
        
        # lag vector
        tt = np.arange(lag_start, lag_end+self.dt, self.dt)
        tp_start = round((lag_start+self.maxlag)/self.dt)
        tp_end = tp_start+len(tt)  # make sure same length, so do not use round((lag_end+self.maxlag+self.dt)/self.dt)

        # select dist_start dist_end
        if source_indx != None:
            a_id = np.where((corr_pair[:,0]==source_indx))[0]
            b_id = np.where((pair_dist>=dist_start) & (pair_dist<=dist_end))[0]
            id = np.intersect1d(a_id ,b_id)
        elif receiver_indx != None:
            a_id = np.where((corr_pair[:,1]==receiver_indx))[0]
            b_id = np.where((pair_dist>=dist_start) & (pair_dist<=dist_end))[0]
            id = np.intersect1d(a_id, b_id)
        else:
            raise ValueError("source_indx or receiver_indx must be specified")
        dist = pair_dist[id]

        # init dist_interval
        if dist_interval == None:
            dist_interval = np.min(dist)
        if dist_interval < np.diff(np.sort(dist)).min():
            raise ValueError("win_interval must be >= (np.diff(np.sort(dist)).min()=%f)" % np.diff(np.sort(dist)).min())
        if dist_interval > np.max(dist):
            raise ValueError("win_interval must be <= (np.max(dist)=%f)" % np.max(dist))
        
        # filter --> data
        if filter:
            data=np.empty((len(id), len(tt)))
            for i in range(0, len(id)):
                data[i,:] = bandpass(self.stack_data[id[i], win_num, tp_start:tp_end], f1, f2, int(1/self.dt), corners=corners, zerophase=zerophase)
        else:
            data = self.stack_data[id, win_num, tp_start:tp_end]
            
        # plot
        colors = list(mcolors.TABLEAU_COLORS.keys())
        if mode == 'waveform':
            if data.size != 0:
                # init fig
                fig, ax = plt.subplots(figsize=figsize)

                # normalize data
                if amp_normalize:
                    scale = np.max(np.abs(data), axis=1)[:,None]
                    if np.min(scale) != 0:
                        data_plot = amp_scale*data/scale
                    else:
                        raise ValueError("data is all zeros in amp_normalize=True")
                else:
                    scale = np.max(np.abs(data))
                    if scale != 0:
                        data_plot = amp_scale*data/scale
                    else:
                        raise ValueError("data is all zeros in amp_normalize=False")
                
                # plot waveform
                for i in range(0, data_plot.shape[0]):
                    ax.plot(tt, amp_scale*data_plot[i]+dist[i], 'k', linewidth=linewidth)

                # plot velocity
                for i in range(0, len(velocity)):
                    x0=0; y0=0
                    if dist_unit == 'm':
                        x1 = (dist_end-dist_start)/velocity[i]
                        y1 = dist_end
                    elif dist_unit == 'km':
                        x1 = (dist_end-dist_start)/1000/velocity[i]
                        y1 = dist_end
                    elif dist_unit == 'degree':
                        x1 = (dist_end-dist_start)/(111.2*1000)/velocity[i]
                        y1 = dist_end
                    else:
                        raise ValueError("dist_unit must be 'm', 'km', or 'degree'")
                    ax.plot([x0,x1], [y0,y1], color=colors[i], linestyle='--', linewidth=1.5, label=str(velocity[i])+'m/s')
                    ax.plot([x0,-x1], [y0,y1], color=colors[i], linestyle='--', linewidth=1.5)

                # set legend
                ax.legend(loc='upper right', fontsize=8, shadow=False)
                ax.set_xlim(tt[0], tt[-1])
                ax.set_xlabel('Time(s)')
                ax.set_ylabel('Distance(%s)' % (dist_unit))

                # set title
                if source_indx != None:
                    ax.set_title("StackData: source_indx=%d, win_num=%d, filter=[%.2f, %.2f] hz" % (source_indx, win_num, f1, f2))
                elif receiver_indx != None:
                    ax.set_title("StackData: receiver_indx=%d, win_num=%d, filter=[%.2f, %.2f] hz" % (source_indx, win_num, f1, f2))
                else:
                    raise ValueError("source_indx or receiver_indx must be specified")
            
                # save or show
                if save:
                    if save_path is None:
                        raise ValueError("save_path must be specified")
                    fig.savefig(os.path.join(save_path), dpi=dpi, format='pdf', bbox_inches = 'tight') 
                    plt.close(fig)
                else:
                    plt.show()
            else:
                raise ValueError("data is empty")

        elif mode=="mat":
            if data.size != 0:
                # init fig
                fig, ax = plt.subplots(figsize=figsize)

                # average waveforms
                ntrace = int((dist_end-dist_start)/dist_interval)
                ndata  = np.zeros(shape=(ntrace,len(tt)))
                for i in range(0,ntrace):
                    tindx = np.where(((dist-dist_start)>=i*dist_interval) & ((dist-dist_start)<(i+1)*dist_interval))[0]
                    ndata[i] = np.mean(data[tindx],axis=0)

                # normalize waveforms
                if amp_normalize:
                    scale = np.max(np.abs(ndata), axis=1)[:,None]
                    if np.min(scale) != 0:
                        ndata_plot = ndata/scale
                    else:
                        raise ValueError("data is all zeros in amp_normalize=True")
                else:
                    scale = np.max(np.abs(ndata))
                    if scale != 0:
                        ndata_plot = ndata/scale
                    else:
                        raise ValueError("data is all zeros in amp_normalize=False")
                
                # plot
                cax=ax.matshow(ndata_plot, extent=[lag_start, lag_end, ntrace, 0], aspect='auto', cmap=cmap) # cmap=cmaps.MPL_jet,
                for i in range(0, len(velocity)):
                    x0=0
                    y0=0
                    if dist_unit == 'm':
                        x1 = (dist_end-dist_start)/velocity[i]
                        y1 = ntrace
                    elif dist_unit == 'km':
                        x1 = (dist_end-dist_start)/1000/velocity[i]
                        y1 = ntrace
                    elif dist_unit == 'degree':
                        x1 = (dist_end-dist_start)/(111.2*1000)/velocity[i]
                        y1 = ntrace
                    else:
                        raise ValueError("dist_unit must be 'm', 'km', or 'degree'")
                    ax.plot([x0,x1], [y0,y1], color=colors[i], linestyle='--', linewidth=1.5, label=str(velocity[i])+'m/s')
                    ax.plot([x0,-x1], [y0,y1], color=colors[i], linestyle='--', linewidth=1.5)

                # set x axis
                ax.legend(loc='upper right', fontsize=7, shadow=False)
                ax.set_xlim(tt[0], tt[-1])
                ax.xaxis.set_ticks_position('bottom')
                ax.set_xlabel('Time(s)')

                # set y axis
                ax.invert_yaxis()
                yy_tick=np.linspace(0, ntrace, num=yticklabel_num)
                yy_label = np.around(np.linspace(dist_start, dist_end, num=yticklabel_num), decimals=2)
                ax.set_yticks(yy_tick)
                ax.set_yticklabels(yy_label)
                ax.set_ylabel('Distance(%s)' % (dist_unit))

                # set title
                if source_indx != None:
                    ax.set_title("StackData: source_indx=%d, win_num=%d, filter=[%.2f, %.2f] hz" % (source_indx, win_num, f1, f2))
                elif receiver_indx != None:
                    ax.set_title("StackData: receiver_indx=%d, win_num=%d, filter=[%.2f, %.2f] hz" % (source_indx, win_num, f1, f2))
                else:
                    raise ValueError("source_indx or receiver_indx must be specified")
                
                # save or show
                if save:
                    if save_path is None:
                        raise ValueError("save_path must be specified")
                    fig.savefig(os.path.join(save_path), dpi=dpi, format='pdf', bbox_inches='tight') 
                    plt.close(fig)
                else:
                    plt.show()
            else:
                raise ValueError("data is empty")
            
        else:
            raise ValueError("mode must be 'waveform' or 'mat'")



    def plot_moveout_all(self, 
                 corr_pair, 
                 pair_dist, 
                 dist_start=None, 
                 dist_end=None, 
                 amp_scale=1,
                 amp_normalize=True, 
                 win_num = 0, 
                 lag_start=None, 
                 lag_end=None, 
                 filter=False, 
                 f1=None, 
                 f2=None,
                 corners=4, 
                 zerophase=True,  
                 dist_interval=None, 
                 mode='waveform', 
                 cmap='seismic', 
                 linewidth=0.8,
                 yticklabel_num=10, 
                 figsize=(10, 6),
                 dist_unit="m", 
                 velocity=[], 
                 save=False, 
                 save_path=None, 
                 dpi=30):      
        
        self.maxlag = self.dt * (self.stack_data.shape[2]-1)/2
        
        # check corr_pair pair_dist
        if corr_pair.shape[0] != len(pair_dist):
            raise ValueError("the length of corr_pair is not consistent with pair_dist")
        
        # check demision of pair_dist
        if pair_dist.ndim == 2:
            pair_dist = pair_dist.reshape(-1)
        
        # check dist_unit
        if dist_unit not in ['m', 'km', 'degree']:
            raise ValueError("dist_unit must be 'm', 'km', or 'degree'")
        
        # init dist_start
        if dist_start == None:
            dist_start = np.min(pair_dist)
        if dist_start < np.min(pair_dist):
            raise ValueError("dist_start must be >= (min(pair_dist)=%f)" % np.min(pair_dist))
        
        # init dist_end
        if dist_end == None:
            dist_end = np.max(pair_dist)
        if dist_end > np.max(pair_dist):
            raise ValueError("dist_end must be <= (max(pair_dist)=%f)" % np.max(pair_dist))

        # init lag_start 
        if lag_start == None:
            lag_start = -self.maxlag
        if lag_start < -self.maxlag:
            raise ValueError("lag_start must be >= (-maxlag=-%f)" % self.maxlag)
        
        # init lag_end
        if lag_end == None:
            lag_end = self.maxlag
        if lag_end > self.maxlag:
            raise ValueError("lag_end must be <= (maxlag=%f)" % self.maxlag)
        
        # lag vector
        tt = np.arange(lag_start, lag_end+self.dt, self.dt)
        tp_start = round((lag_start+self.maxlag)/self.dt)
        tp_end = tp_start+len(tt)  # make sure same length, so do not use round((lag_end+self.maxlag+self.dt)/self.dt)

        # select dist_start dist_end
        id = np.where((pair_dist>=dist_start) & (pair_dist<=dist_end))[0]
        dist = pair_dist[id]

        # init dist_interval
        if dist_interval == None:
            dist_interval = np.min(dist)
        if dist_interval < np.diff(np.sort(dist)).min():
            raise ValueError("win_interval must be >= (np.diff(np.sort(dist)).min()=%f)" % np.diff(np.sort(dist)).min())
        if dist_interval > np.max(dist):
            raise ValueError("win_interval must be <= (np.max(dist)=%f)" % np.max(dist))
        
        # filter --> data
        if filter:
            data=np.empty((len(id), len(tt)))
            for i in range(0, len(id)):
                data[i,:] = bandpass(self.stack_data[id[i], win_num, tp_start:tp_end], f1, f2, int(1/self.dt), corners=corners, zerophase=zerophase)
        else:
            data = self.stack_data[id, win_num, tp_start:tp_end]
            
        # plot
        colors = list(mcolors.TABLEAU_COLORS.keys())
        if mode == 'waveform':
            if data.size != 0:
                # init fig
                fig, ax = plt.subplots(figsize=figsize)

                # normalize data
                if amp_normalize:
                    scale = np.max(np.abs(data), axis=1)[:,None]
                    if np.min(scale) != 0:
                        data_plot = amp_scale*data/scale
                    else:
                        raise ValueError("data is all zeros in amp_normalize=True")
                else:
                    scale = np.max(np.abs(data))
                    if scale != 0:
                        data_plot = amp_scale*data/scale
                    else:
                        raise ValueError("data is all zeros in amp_normalize=False")
                
                # plot waveform
                for i in range(0, data_plot.shape[0]):
                    ax.plot(tt, amp_scale*data_plot[i]+dist[i], 'k', linewidth=linewidth)

                # plot velocity
                for i in range(0, len(velocity)):
                    x0=0; y0=0
                    if dist_unit == 'm':
                        x1 = (dist_end-dist_start)/velocity[i]
                        y1 = dist_end
                    elif dist_unit == 'km':
                        x1 = (dist_end-dist_start)/1000/velocity[i]
                        y1 = dist_end
                    elif dist_unit == 'degree':
                        x1 = (dist_end-dist_start)/(111.2*1000)/velocity[i]
                        y1 = dist_end
                    else:
                        raise ValueError("dist_unit must be 'm', 'km', or 'degree'")
                    ax.plot([x0,x1], [y0,y1], color=colors[i], linestyle='--', linewidth=1.5, label=str(velocity[i])+'m/s')
                    ax.plot([x0,-x1], [y0,y1], color=colors[i], linestyle='--', linewidth=1.5)

                # set axis
                ax.set_xlim(tt[0], tt[-1])
                ax.set_xlabel('Time(s)')
                ax.set_ylabel('Distance(%s)' % (dist_unit))
                ax.legend(loc='upper right', fontsize=8, shadow=False)
                ax.set_title("StackData: win_num=%d, filter=[%.2f, %.2f] hz" % (win_num, f1, f2))
            
                # save or show
                if save:
                    if save_path is None:
                        raise ValueError("save_path must be specified")
                    fig.savefig(os.path.join(save_path), dpi=dpi, format='pdf', bbox_inches = 'tight') 
                    plt.close(fig)
                else:
                    plt.show()

            else:
                raise ValueError("data is empty")

        elif mode=="mat":
            if data.size != 0:
                # init fig
                fig, ax = plt.subplots(figsize=figsize)

                # average waveforms
                ntrace = int((dist_end-dist_start)/dist_interval)
                ndata  = np.zeros(shape=(ntrace,len(tt)))
                for i in range(0,ntrace):
                    tindx = np.where(((dist-dist_start)>=i*dist_interval) & ((dist-dist_start)<(i+1)*dist_interval))[0]
                    ndata[i] = np.mean(data[tindx],axis=0)

                # normalize waveforms
                if amp_normalize:
                    scale = np.max(np.abs(ndata), axis=1)[:,None]
                    if np.min(scale) != 0:
                        ndata_plot = ndata/scale
                    else:
                        raise ValueError("data is all zeros in amp_normalize=True")
                else:
                    scale = np.max(np.abs(ndata))
                    if scale != 0:
                        ndata_plot = ndata/scale
                    else:
                        raise ValueError("data is all zeros in amp_normalize=False")
                
                # plot
                cax=ax.matshow(ndata_plot, extent=[lag_start, lag_end, ntrace, 0], aspect='auto', cmap=cmap) # cmap=cmaps.MPL_jet,
                for i in range(0, len(velocity)):
                    x0=0
                    y0=0
                    if dist_unit == 'm':
                        x1 = (dist_end-dist_start)/velocity[i]
                        y1 = ntrace
                    elif dist_unit == 'km':
                        x1 = (dist_end-dist_start)/1000/velocity[i]
                        y1 = ntrace
                    elif dist_unit == 'degree':
                        x1 = (dist_end-dist_start)/(111.2*1000)/velocity[i]
                        y1 = ntrace
                    else:
                        raise ValueError("dist_unit must be 'm', 'km', or 'degree'")
                    ax.plot([x0,x1], [y0,y1], color=colors[i], linestyle='--', linewidth=1.5, label=str(velocity[i])+'m/s')
                    ax.plot([x0,-x1], [y0,y1], color=colors[i], linestyle='--', linewidth=1.5)

                # set x axis
                ax.set_title("StackData: win_num=%d, filter=[%.2f, %.2f] hz" % (win_num, f1, f2))
                ax.legend(loc='upper right', fontsize=7, shadow=False)
                ax.set_xlim(tt[0], tt[-1])
                ax.xaxis.set_ticks_position('bottom')
                ax.set_xlabel('Time(s)')

                # set y axis
                ax.invert_yaxis()
                yy_tick=np.linspace(0, ntrace, num=yticklabel_num)
                yy_label = np.around(np.linspace(dist_start, dist_end, num=yticklabel_num), decimals=2)
                ax.set_yticks(yy_tick)
                ax.set_yticklabels(yy_label)
                ax.set_ylabel('Distance(%s)' % (dist_unit))

                # save or show
                if save:
                    if save_path is None:
                        raise ValueError("save_path must be specified")
                    fig.savefig(os.path.join(save_path), dpi=dpi, format='pdf', bbox_inches='tight') 
                    plt.close(fig)
                else:
                    plt.show()
            else:
                raise ValueError("data is empty")
            
        else:
            raise ValueError("mode must be 'waveform' or 'mat'")





import os
import re
import json
import numpy as np

from noiseflow.cc.python.rfft import RFFTClass_python
from noiseflow.cc.python.corr import CorrClass_python
from noiseflow.cc.python.stack import StackClass_python

from noiseflow.config.rfftdata import RFFTData_Class
from noiseflow.config.corrdata import CorrData_Class
from noiseflow.config.stackdata import StackData_Class


# Determine the absolute path to the config file
env = os.environ.get('CONDA_DEFAULT_ENV')

if env is None:
    env = ''
else:
    env = re.sub('[^a-zA-Z0-9_]', '', env)[0:50]
    
config_path = os.path.abspath(os.path.expanduser(f'~/.noiseflow/config_{env}.json'))    

# Read the config.json file
with open(config_path) as f:
    config = json.load(f)

if config.get('NOISEFLOW_USE_CPP', False):
    NOISEFLOW_USE_CPP = True
    from noiseflow.lib import cc_share
else:
    NOISEFLOW_USE_CPP = False  


def rfft(raw_data, dt,
        cc_len, cc_step, 
        time_norm, clip_std, smooth_N,
        freq_norm, freqmin, freqmax, whiten_npad, smoothspect_N, 
        flag=False, flag_gap=None, threads=1, jobs=1, py=False):
    
    # check demision of raw_data
    if raw_data.ndim == 1:
        raw_data = raw_data.reshape(1, -1)

    # check freqmax
    if (freqmax > (1/(2*dt))):
        raise ValueError("freqmax must be <= (%fhz) according to fmax=1/(2*dt) formula" % (1/(2*dt)))
    
    # check cc_len
    if (cc_len > raw_data.shape[1]*dt):
        raise ValueError("cc_len must be <= (raw_data.shape[1]*dt=%f)" % (raw_data.shape[1]*dt))

    # check cc_step
    if (cc_step >= cc_len):
        raise ValueError("cc_step must be < (cc_len=%f)" % cc_len)

    # check time_norm
    if time_norm not in ['no', 'onebit', 'clip', 'smooth']:
        raise ValueError("time_norm must be 'no', 'onebit', 'clip', or 'smooth'")
    
    # check freq_norm
    if freq_norm not in ['no', 'whiten', 'smooth_whiten']:
        raise ValueError("freq_norm must be 'no', 'whiten', or 'smooth_whiten'")
    
    # check threads
    if threads > raw_data.shape[0]:
        raise ValueError("threads must be <= (raw_data.shape[0]=%d)" % raw_data.shape[0])
    
    # check flag_gap
    if flag_gap==None:
        flag_gap = int(raw_data.shape[0])

    # run rfft
    if py:
        if raw_data.dtype == np.dtype(np.float32) or raw_data.dtype == np.dtype(np.float64):
            m = RFFTClass_python(raw_data, dt,
                         cc_len, cc_step, 
                        time_norm, clip_std, smooth_N,
                        freq_norm, freqmin, freqmax, whiten_npad, smoothspect_N,
                        flag, jobs)
            m.run()
            rfft_data = m.output_data
        else:
            raise TypeError("raw_data.dtype must be np.float32 or np.float64")
    else:
        if raw_data.dtype == np.dtype(np.float32) and NOISEFLOW_USE_CPP:
            rfft_data = cc_share.rfft_float(raw_data, dt, 
                cc_len, cc_step, 
                time_norm, clip_std, smooth_N,
                freq_norm, freqmin, freqmax, whiten_npad, smoothspect_N,
                flag, flag_gap,
                threads)
        elif raw_data.dtype == np.dtype(np.float64) and NOISEFLOW_USE_CPP:
            rfft_data = cc_share.rfft_double(raw_data, dt, 
                cc_len, cc_step, 
                time_norm, clip_std, smooth_N,
                freq_norm, freqmin, freqmax, whiten_npad, smoothspect_N,
                flag, flag_gap,
                threads)
        else:
            raise TypeError("raw_data.dtype must be np.float32, np.float64 or NOISEFLOW_USE_CPP=False")
    
    # generate RFFTData object
    RFFTData = RFFTData_Class(rfft_data, dt, cc_len, cc_step, 
        time_norm, clip_std, smooth_N,
        freq_norm, freqmin, freqmax, whiten_npad, smoothspect_N,
        flag, flag_gap,
        threads, jobs, py)
    
    return RFFTData



def corr(rfft_data, dt, corr_method, corr_pair, maxlag, smoothspect_N=10, 
         flag=False, flag_gap=None, threads=1, jobs=1, py=False):
    
    # check corr_method
    if corr_method not in ['xcorr', 'deconv', 'coherency']:
        raise ValueError("corr_method must be 'xcorr', 'deconv', or 'coherency'")
    
    # check maxlag
    if maxlag > dt*rfft_data.shape[2]:
        raise ValueError("maxlag must be <= (dt*rfft_data.shape[2]=%f)" % (dt*rfft_data.shape[2]))
    
    # check threads
    if threads > rfft_data.shape[0]:
        raise ValueError("threads must be <= (rfft_data.shape[0]=%d)" % rfft_data.shape[0])

    # check demision of corr_pair
    if corr_pair.ndim == 1:
        corr_pair = corr_pair.reshape(1, -1)

    # check threads
    if threads > corr_pair.shape[0]:
        raise ValueError("threads must be <= (corr_pair.shape[0]=%d)" % corr_pair.shape[0])
    
    # check flag_gap
    if flag_gap==None:
        flag_gap = int(corr_pair.shape[0])

    # run corr
    if py:
        if rfft_data.dtype == np.dtype(np.complex64) or rfft_data.dtype == np.dtype(np.complex128):
            m = CorrClass_python(rfft_data, dt, corr_method, corr_pair, maxlag,
                smoothspect_N, flag, jobs)
            m.run()
            corr_data = m.output_data
        else:
            raise TypeError("rfft_data.dtype must be np.complex64 or np.complex128")
    else:
        if rfft_data.dtype == np.dtype(np.complex64) and NOISEFLOW_USE_CPP:
            corr_data = cc_share.corr_float(rfft_data, dt, corr_method, corr_pair, maxlag,
                smoothspect_N, flag, flag_gap, threads)
        elif rfft_data.dtype == np.dtype(np.complex128) and NOISEFLOW_USE_CPP:
            corr_data = cc_share.corr_double(rfft_data, dt, corr_method, corr_pair, maxlag,
                smoothspect_N, flag, flag_gap, threads)
        else:
            raise TypeError("rfft_data.dtype must be np.complex64, np.complex128 or NOISEFLOW_USE_CPP=False")

    # generate CorrData object
    CorrData = CorrData_Class(corr_data, dt, corr_method, corr_pair, maxlag, smoothspect_N,
        flag, flag_gap, threads, jobs, py)
    
    return CorrData



def stack(corr_data, dt, stack_method, 
          par=None, 
          stack_all=True, stack_len=0, stack_step=0, 
          pick=False, median_high=10, median_low=0.1, 
          flag=False, flag_gap=None, threads=1, jobs=1, py=False):
    
    # check par
    if par == None:
        par = {"axis":0,"p":2,"g":1,"cc_min":0.0,"epsilon":1E-5,"maxstep":10,
                "win":None,"stat":False,"h":0.75,'plot':False,'normalize':True,'ref':None},
    
    # check demision of corr_data
    if corr_data.ndim == 2:
        corr_data = corr_data.reshape(1, corr_data.shape[0], corr_data.shape[1])

    # check stack_method
    if stack_method not in ["linear","pws","robust","acf","nroot","selective","cluster","tfpws"]:
        raise ValueError("stack_method must be 'linear', 'pws', 'robust', 'acf', 'nroot', 'selective', 'cluster', or 'tfpws'")
    
    # check stack_len
    if (stack_len > corr_data.shape[1]):
        raise ValueError("stack_len must be <= (corr_data.shape[1]=%d)" % corr_data.shape[1])

    # check stack_step
    if (stack_step >= stack_len):
        raise ValueError("stack_step must be < (stack_len=%d)" % stack_len)
    
    # check threads
    if threads > corr_data.shape[0]:
        raise ValueError("threads must be <= (corr_data.shape[0]=%d)" % corr_data.shape[0])

    # check flag_gap
    if flag_gap==None:
        flag_gap = int(corr_data.shape[0])

    # run stack
    if py:
        if corr_data.dtype == np.dtype(np.float32) or corr_data.dtype == np.dtype(np.float64):
            m = StackClass_python(corr_data, stack_method, par, stack_all, stack_len, stack_step, pick, median_high, median_low, flag, jobs)
            m.run()
            stack_data = m.output_data
            stack_ngood = m.ngood_all
        else:
            raise TypeError("corr_data.dtype must be np.float32 or np.float64")
    else:
        if corr_data.dtype == np.dtype(np.float32) and NOISEFLOW_USE_CPP:
            m = cc_share.StackClass_float(corr_data, stack_method, stack_all, stack_len, stack_step, pick, median_high, median_low, flag, flag_gap, threads)
            m.run()
            stack_data = m.get_stackdata()
            stack_ngood = m.get_ngood()
        elif corr_data.dtype == np.dtype(np.float64) and NOISEFLOW_USE_CPP:
            m = cc_share.StackClass_double(corr_data, stack_method, stack_all, stack_len, stack_step, pick, median_high, median_low, flag, flag_gap, threads)
            m.run()
            stack_data = m.get_stackdata()
            stack_ngood = m.get_ngood()
        else:
            raise TypeError("corr_data.dtype must be np.float32, np.float64 or NOISEFLOW_USE_CPP=False")
    
    # generate StackData object
    StackData = StackData_Class(stack_data, stack_ngood, dt,
                 stack_method, par, stack_all, stack_len, stack_step, pick, median_high, median_low, flag, flag_gap, threads, jobs, py)
    
    return StackData






import time
import scipy
import numpy as np

from tqdm import tqdm
from joblib import Parallel, delayed
from noiseflow.cc.python.utils import slice_window, moving_ave, whiten, split


def time_norm_clip(data, clip_std):
    lim = clip_std * np.std(data)
    time_norm_data = np.clip(data, -lim, lim)

    return time_norm_data


def time_norm_smooth(data, smooth_N):
    temp = moving_ave(np.abs(data), smooth_N)
    time_norm_data = np.divide(data, 
                               temp, 
                               out=np.zeros_like(data, dtype=data.dtype), 
                               where=temp!=0)

    return time_norm_data


class RFFTClass_python(object):
    def __init__(self, raw_data, dt, 
                cc_len, cc_step, 
                time_norm, clip_std, smooth_N,
                freq_norm, freqmin, freqmax, whiten_npad, smoothspect_N,
                flag, jobs):
        
        self.raw_data = raw_data
        self.dt = dt
        self.cc_len = cc_len
        self.cc_step = cc_step
        self.time_norm_method = time_norm
        self.clip_std = clip_std
        self.smooth_N = smooth_N
        self.freq_norm_method = freq_norm
        self.freqmin = freqmin
        self.freqmax = freqmax
        self.whiten_npad = whiten_npad
        self.smoothspect_N = smoothspect_N
        self.flag = flag
        self.jobs = jobs

        # convert time to points
        self.cc_len_points = int(cc_len/dt)
        self.cc_step_points = int(cc_step/dt)
        self.npts = int(raw_data.shape[1])
        self.rfft_npts = int(self.cc_len_points/2+1)
        self.channel_num = int(raw_data.shape[0])
  
        # slice_window --> _win_num, _win_info
        self.win_info = slice_window(self.npts, self.cc_len_points, self.cc_step_points)
        self.win_num = int(self.win_info.shape[0])

        # initialize output_data
        if self.raw_data.dtype == np.dtype(np.float32):
            self.output_data = np.empty((self.channel_num, self.win_num, self.rfft_npts), dtype=np.complex64)
            self.time_dtype = np.float32
            self.freq_dtype = np.complex64
        elif self.raw_data.dtype == np.dtype(np.float64):
            self.output_data = np.empty((self.channel_num, self.win_num, self.rfft_npts), dtype=np.complex128)
            self.time_dtype = np.float64
            self.freq_dtype = np.complex128
        else:
            print("error: please input the correct data type.")
            exit(1)


    def time_norm(self, data):
        time_norm_data = None
        if self.time_norm_method == "no":
            time_norm_data = data
        elif self.time_norm_method == "clip":
            time_norm_data = time_norm_clip(data, self.clip_std)
        elif self.time_norm_method == "onebit":
            time_norm_data = np.sign(data)
        elif self.time_norm_method == "smooth":
            time_norm_data = time_norm_smooth(data, self.smooth_N)
        else:
            print("error: please input the correct time_norm method.")
            exit(1)

        return time_norm_data
    

    def rfft(self, time_norm_data):
        rfft_data = np.empty((self.win_num, self.rfft_npts), dtype=self.freq_dtype)

        for i in range(self.win_num):
            segment = time_norm_data[self.win_info[i, 0]:self.win_info[i, 1]]
            segment_rfft = scipy.fft.rfft(segment)
            rfft_data[i] = segment_rfft

        return rfft_data


    def freq_norm(self, rfft_data):
        freq_norm_data = None
        if self.freq_norm_method == "no":
            freq_norm_data = rfft_data
        else:
            freq_norm_data = whiten(rfft_data, self.dt, self.freq_norm_method, self.freqmin, self.freqmax, self.smoothspect_N, self.whiten_npad)

        return freq_norm_data


    def process_chunk(self, chunk_start, chunk_end):
        results = []
        if self.flag and chunk_start == 0:
            bar = tqdm(range(chunk_start, chunk_end))
        else:
            bar = range(chunk_start, chunk_end)

        for i in bar:
            # time_norm
            time_norm_data = self.time_norm(self.raw_data[i, :])

            # rfft
            rfft_data = self.rfft(time_norm_data)

            # freq_norm
            rfft_norm_data = self.freq_norm(rfft_data)

            results.append(rfft_norm_data)

        return results


    def run(self):
        if self.flag:
            start_time = time.time()
            print(f"Start rfft with {self.jobs} jobs in python...")
        
        # parallel processing
        if self.jobs > 1:
            chunk_list = split(num=self.channel_num, n_jobs=self.jobs)
            results = Parallel(n_jobs=self.jobs, backend="loky")(delayed(self.process_chunk)(chunk_start, chunk_end) for chunk_start, chunk_end in chunk_list)

            for i, chunk_results in enumerate(results):
                chunk_start = chunk_list[i][0]
                for j, result in enumerate(chunk_results):
                    self.output_data[chunk_start + j] = result
        else:
            results = self.process_chunk(0, self.channel_num)
            self.output_data = np.array(results, dtype=self.freq_dtype)

        if self.flag:
            end_time = time.time()
            elapsed_time = end_time - start_time
            print(f"End rfft with total time {elapsed_time}s")




import os
import re
import json


env = os.environ.get('CONDA_DEFAULT_ENV')

if env is None:
    env = ''
else:
    env = re.sub('[^a-zA-Z0-9_]', '', env)[0:50]
    
config_path = os.path.abspath(os.path.expanduser(f'~/.noiseflow/config_{env}.json'))    

try:
    from noiseflow.lib import cc_share
    from noiseflow.lib import signal_share
    NOISEFLOW_USE_CPP = True
except:
    NOISEFLOW_USE_CPP = False

os.makedirs(os.path.expanduser('~/.noiseflow'), exist_ok=True)
compile_time_env = {"NOISEFLOW_USE_CPP": NOISEFLOW_USE_CPP}
with open(config_path, 'w') as f:
    json.dump(compile_time_env, f)


from noiseflow.config.rawdata import RawData_Class
from noiseflow.config.rfftdata import RFFTData_Class
from noiseflow.config.corrdata import CorrData_Class
from noiseflow.config.stackdata import StackData_Class

from noiseflow.cc.wrapper import rfft, corr, stack
from noiseflow.client.client import downloader
from noiseflow.signal.wrapper import bandpass, bandstop, lowpass, highpass, detrend, decimate, taper

from noiseflow.utils.load import load_raw, load_rfft, load_corr, load_stack
from noiseflow.utils.timestamp import time_linspace, get_timestamp, get_stack_timestamp


__all__ = ["cc",
           "client",
           "config",
           "signal",
           "utils",
           "tests"]

__version__ = "0.0.1"



import os
import requests

from tqdm import tqdm
from faker import Faker
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor, as_completed


cpu_cores = os.cpu_count()
requests.packages.urllib3.disable_warnings()


def downloader(url, type='https', show_info = True, resume=True, filename=None, num_threads=cpu_cores, timeout=10, chunk_size=1024*1000, header=None, proxies=None):
    if type == 'https':
        d = downloader_https(url, show_info, resume, filename, num_threads, timeout, chunk_size, header, proxies)
        d.download()
    else:
        pass


class downloader_https():
    def __init__(self, url, show_info = True, resume=True, filename=None, num_threads=cpu_cores, timeout=10, chunk_size=1024*1000, header=None, proxies=None):
        """
        :param url: link address
        :param filename: file name
        """
        self.url = url
        self.show_info = show_info
        self.resume = resume
        self.chunk_size = chunk_size 
        self.filename = filename
        self.num_threads = num_threads
        self.proxies = proxies
        self.timeout = timeout
        self.file_type = None
        self.accept_ranges = None
        self.content_length = None
        self.transfer_encoding = None
        if header is None:
            self.header = {}
            self.header.setdefault('User-Agent', Faker().user_agent())
        elif 'User-Agent' not in header:
            self.header.setdefault('User-Agent', Faker().user_agent())
        else:
            self.header = header

 
    def check_url(self):
        """
        check url support break-point resume and support multi-thread downloading
        """
        _, filename = os.path.split(self.url)
        self.filename = self.filename or filename

        res = requests.head(self.url, headers=self.header, proxies=self.proxies, timeout=self.timeout, allow_redirects=True, verify=False)  # verify=False 关闭ssl双向验证，解决访问https报错问题

        if not (200 <= res.status_code < 400):
            raise Exception('Bad request!')

        headers = res.headers
        self.file_type = headers.get('Content-Type')
        self.accept_ranges = headers.get('Accept-Ranges')
        self.transfer_encoding = headers.get('Transfer-Encoding')

        if self.transfer_encoding == "chunked" or self.transfer_encoding == "gzip, chunked":
            self.num_threads = 1
            self.content_length = 0
        else:
            lengths = headers.get('Content-Length')
            if lengths == None:
                self.content_length = 0
            else:
                self.content_length = int(lengths)


    def get_range(self, start=0):
        """
        set download range
        eg: [(0, 1023), (1024, 2047), (2048, 3071) ...]
        """
        if self.transfer_encoding == "chunked" or self.transfer_encoding == "gzip, chunked":
            _range = [(start, '')]
        else:
            lst = range(start, self.content_length, self.chunk_size)   
            _range = list(zip(lst[:-1], [i - 1 for i in lst[1:]]))
            _range.append((lst[-1], ''))

        return _range


    def download_by_piece(self, _range):
        start, stop = _range
        headers = {**self.header, **{"Range": f"bytes={start}-{stop}"}} # merge

        res = requests.get(self.url, headers=headers, proxies=self.proxies, timeout=self.timeout, allow_redirects=True, verify=False)
        if res.status_code != 206:
            raise Exception(f'Request raise error, url: {self.url}, range: {_range}')
        return _range, res.content


    def download(self):
        start = 0
        self.check_url()

        if self.accept_ranges != "bytes":
            if self.show_info:
                print(f'--- Mission ---: {self.url} download from scratch || with single thread, do not support breakpoint resuming')
            
            file_path = Path(self.filename)

            res = requests.get(self.url, 
                                headers=self.header, 
                                proxies=self.proxies, 
                                timeout=self.timeout, 
                                allow_redirects=True, 
                                verify=False)

            if res.status_code != 206:
                raise Exception(f'Request raise error, url: {self.url}')
 
            # write
            open(file_path, 'w').close() 
            with open(self.filename, 'rb+') as fp:
                fp.seek(0)
                fp.write(res.content)

            if self.show_info:
                print(f'--- File ---: {self.filename} download completely')
        else:
            file_path = Path(self.filename)

            if self.resume:
                open(file_path, 'w+').close()
                start = 0
                if self.show_info:
                    print(f'--- Mission ---: {self.url} download from scratch || with {self.num_threads} threads, support breakpoint resuming')

            else:
                if file_path.exists():
                    # breakpoint resume
                    start = file_path.lstat().st_size
                    if self.show_info:
                        print(f'--- Mission ---: {self.url} download from breakpoint || with {self.num_threads} threads, support breakpoint resuming')

                    # If file have already downloaded 
                    if start == self.content_length:
                        if self.show_info:
                            print(f'--- File ---: {self.filename} has already been downloaded completely')
                        return
                else:
                    open(file_path, 'w+').close()
                    start = 0
                    if self.show_info:
                        print(f'--- Mission ---: {self.url} download from scratch || with {self.num_threads} threads, support breakpoint resuming')

            # init progress bar
            if self.show_info:
                pbar = tqdm(total=self.content_length,
                        initial=start,
                        unit='B',
                        unit_scale=True,
                        desc=self.filename,
                        unit_divisor=1024)
            
            # multi-thread download
            with ThreadPoolExecutor(max_workers=self.num_threads) as pool:
                res = [pool.submit(self.download_by_piece, r) for r in self.get_range(start=start)] # or use map function

                # write
                with open(self.filename, 'rb+') as fp:
                    for item in as_completed(res):
                        _range, content = item.result()
                        start, stop = _range
                        fp.seek(start)
                        fp.write(content)
                        # update progress bar
                        if self.show_info:
                            pbar.update(self.chunk_size)

            if self.show_info:
                pbar.close()
                print(f'--- File ---: {self.filename} download completely')


    def print(self):
        self.check_url()
        print( "self.url = ", self.url, '\n',
            "self.resume = ", self.resume, '\n',
            "self.chunk_size = ", self.chunk_size, '\n',
            "self.filename = ", self.filename, '\n',
            "self.num_threads = ", self.num_threads, '\n',
            "self.proxies = ", self.proxies, '\n',
            "self.timeout = ", self.timeout, '\n',
            "self.file_type = ", self.file_type, '\n',
            "self.accept_ranges = ", self.accept_ranges, '\n',
            "self.content_length = ", self.content_length, '\n',
            "self.transfer_encoding = ", self.transfer_encoding, '\n',
            "self.header = ", self.header
            )




#%%
import h5py
import numpy as np
import matplotlib.pyplot as plt

from scipy.io import savemat
from noiseflow.signal.wrapper import bandpass, bandstop, lowpass, highpass, detrend, decimate, taper


class RawData_Class(object):
    def __init__(self, data, sampling_rate):
        self.data = data
        self.sampling_rate = sampling_rate


    def detrend(self, type='linear', flag=False, flag_gap=None, threads=1, py=False):
        detrend(self.data, type, flag, flag_gap, threads, py)


    def taper(self, max_percentage=0.05, type='hann', side='both', flag=False, flag_gap=None, threads=1, py=False):  
        taper(self.data, max_percentage, type, side, flag, flag_gap, threads, py)


    def bandpass(self, freqmin, freqmax, corners=4, zerophase=True, flag=False, flag_gap=None, threads=1, py=False):
        bandpass(self.data, freqmin, freqmax, self.sampling_rate, corners, zerophase, flag, flag_gap, threads, py)


    def bandstop(self, freqmin, freqmax, corners=4, zerophase=True, flag=False, flag_gap=None, threads=1, py=False):  
        bandstop(self.data, freqmin, freqmax, self.sampling_rate, corners, zerophase, flag, flag_gap, threads, py)
        

    def lowpass(self, freq, corners=4, zerophase=True, flag=False, flag_gap=None, threads=1, py=False):
        lowpass(self.data, freq, self.sampling_rate, corners, zerophase, flag, flag_gap, threads, py)
        

    def highpass(self, freq, corners=4, zerophase=True, flag=False, flag_gap=None, threads=1, py=False):
        highpass(self.data, freq, self.sampling_rate, corners, zerophase, flag, flag_gap, threads, py)
        

    def decimate(self, sampling_rate_new, flag=False, flag_gap=None, threads=1, py=False):
        results = decimate(self.data, self.sampling_rate, sampling_rate_new, flag, flag_gap, threads, py)
        self.data = results
        self.sampling_rate = sampling_rate_new


    # save data
    def save(self, save_path, format='npz', compression=False, h5_compression_format='gzip', h5_compression_opts=3):
        if format == 'npz':
            if compression:
                np.savez_compressed(save_path,
                                    data = self.data,
                                    sampling_rate = self.sampling_rate)
            else:
                np.savez(save_path,
                        data = self.data,
                        sampling_rate = self.sampling_rate)
                
        elif format == 'h5':
            with h5py.File(save_path, 'w') as f:
                group = f.create_group('noiseflow_group')
                group.attrs['sampling_rate'] = self.sampling_rate
                if compression:
                    group.create_dataset('data', data=self.data, compression=h5_compression_format, compression_opts=h5_compression_opts)
                else:
                    group.create_dataset('data', data=self.data)

        elif format == 'mat':
            savemat(save_path, 
                {'data': self.data,
                'sampling_rate': self.sampling_rate},
                do_compression=compression)
            
        else:
            raise ValueError('format must be npz, h5, or mat')


    def plot(self, channel_num=0, win_num=0, save=False, save_path=None, dpi=100):
        pass





# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['noiseflow',
 'noiseflow.cc',
 'noiseflow.cc.python',
 'noiseflow.client',
 'noiseflow.config',
 'noiseflow.signal',
 'noiseflow.signal.python',
 'noiseflow.tests',
 'noiseflow.tests.test_cc',
 'noiseflow.tests.test_cc.test_python',
 'noiseflow.tests.test_client',
 'noiseflow.tests.test_config',
 'noiseflow.tests.test_signal',
 'noiseflow.tests.test_signal.test_python',
 'noiseflow.tests.test_utils',
 'noiseflow.utils']

package_data = \
{'': ['*'],
 'noiseflow.cc': ['include/*', 'src/*'],
 'noiseflow.signal': ['include/*', 'src/*']}

install_requires = \
['faker>=18.13.0,<19.0.0',
 'h5py>=3.7.0,<4.0.0',
 'joblib>=1.3.0,<2.0.0',
 'matplotlib>=3.7.1,<4.0.0',
 'numpy>=1.25.0,<2.0.0',
 'obspy>=1.4.0,<2.0.0',
 'scipy>=1.10.1,<2.0.0',
 'tqdm>=4.65.0,<5.0.0']

setup_kwargs = {
    'name': 'noiseflow',
    'version': '0.0.8',
    'description': 'An ambient noise package',
    'long_description': '# NoiseFlow\n\n\n[![Actions Status](https://github.com/shakeflow/noiseflow/actions/workflows/workflow.yml/badge.svg)](https://github.com/shakeflow/noiseflow/actions)\n[![coverage](https://codecov.io/gh/shakeflow/noiseflow/branch/main/graph/badge.svg)](https://codecov.io/gh/shakeflow/noiseflow)\n[![docs](https://img.shields.io/badge/docs-stable-blue.svg)](https://shakeflow.github.io/noiseflow/)\n[![supported versions](https://img.shields.io/pypi/pyversions/noiseflow.svg?label=python_versions)](https://pypi.python.org/pypi/noiseflow)\n[![docs](https://badge.fury.io/py/noiseflow.svg)](https://badge.fury.io/py/noiseflow)\n\n\n## Prerequisites\n\nNoiseFlow now supports `Clang` and `GCC` compiler in MacOS and Linux system separately, and all installation processes are under the `conda` environment, and we recommend to use miniconda. Make sure to install the following pre-packages before installing noiseflow:\n\n\nIf you use `Clang` in Mac, please install `OpenMP` via `brew` as following:\n\n```bash\nbrew install openmp\n```\n\nAnd use `pip` and `conda` to install the following packages:\n\n```bash\npip install joblib\n\nconda install -c conda-forge numpy scipy matplotlib \nconda install -c conda-forge obspy\nconda install -c conda-forge fftw (动态库)\nconda install -c conda-forge pybind11 (头文件)\nconda install -c conda-forge xtensor xsimd xtl xtensor-blas xtensor-python (可能是静态库)\nconda install -c conda-forge xtensor-fftw  #(usually failed at most time)  \n```\n\nThe `xtensor-fftw` and `KFR` need to be installed from source, first download them:\n\n\n```bash\ngit clone https://github.com/OUCyf/noiseflow.git\ncd noiseflow\ngit submodule init\ngit submodule update\n```\n\n\n\nNote the `xtensor-fftw` do not support M1 chip, and if it is failed to install via conda, you can install it from source into conda environment as `$CONDA_PREFIX`\n\n```bash\ncd ./external/xtensor-fftw\nmkdir build && cd build\ncmake .. -DCMAKE_INSTALL_PREFIX=$CONDA_PREFIX\nmake install\n```\n\n\n\nThe `KFR` package is C++ DSP framework, should be installed in `./extern/kfr` from source\n\n```bash\ncd ./external/kfr\nmkdir build && cd build\ncmake .. -DCMAKE_INSTALL_PREFIX=$CONDA_PREFIX\nmake install\n```\n\n\n\n\n## Installation\n\nNow you can install `NoiseFlow`. If you use `MacOS`, please make sure to use Clang as the complier\n\n```bash\nexport CXX=clang++\n# unset CXX\npython setup.py install\n```\n\nIf you use `Linux`, please use GCC as the compiler\n\n```bash\nexport CXX=g++-13\npython setup.py install\n```\n\n\nIf you use `HPC` with `module` tool, you can use both Clang and GCC, for example using NOTS in Rice University.\n\n```bash\n# use gcc\nmodule load GCC/13.1.0\nexport CXX=g++\npython setup.py install\nINCLUDE_CMAKE_EXTENSION=1 pip install .\n\n# use clang\nmodule load GCCcore/11.2.0\nmodule load Clang/13.0.1\nexport CXX=clang++\npython setup.py install\n```\n\n```bash\nconda install -c conda-forge stockwell\n\nNOISEFLOW_USE_CPP=1 pip install --no-binary :all: noiseflow --no-cache-dir\n\ngit submodule add https://gitclone.com/github.com/kfrlib/kfr.git extern/kfr\n```\nw\n\n\n## License\nNoiseflow is dual-licensed, available under both commercial and apache license.\n\nIf you want to use noiseflow in a commercial product or a closed-source project, you need to purchase a Commercial License.\n\n\n\n`stockwell`: \'numpy>=1.18\', scipy, fftw, [\'cp36-*\', \'cp37-*\', \'cp38-*\', \'cp39-*\', \'cp310-*\']\n\n\n\n## INSTALL (NOETS)\n- conda:\n\n\n\n```bash\nbrew install fftw\nconda install -c conda-forge fftw(没用)\nconda install -c conda-forge pybind11 (头文件)\nconda install -c conda-forge xtensor xsimd xtl xtensor-blas xtensor-python (可能是静态库)\n```\n\n\n\n\n```bash\ncd ./external/xtensor-fftw\nmkdir build && cd build\n\nmyenv = $(conda env list | grep \' \\* \' | awk \'{print $1}\')\nconda_prefix = $(conda info --base)\nexport CMAKE_PREFIX_PATH="${conda_prefix}/envs/${myenv}"\n\n\ncmake .. -DCMAKE_INSTALL_PREFIX=$CONDA_PREFIX -DCMAKE_PREFIX_PATH=$CONDA_PREFIX\nmake install\n```\n\n\n\n### [build-system]\ncmake = \'^3.18.0\'\nfftw = \'^3.3.10\'\npybind11 = \'^2.10.4\'\nxtensor = \'^0.24.6\'\nxsimd = \'8.0.5\' # when the version is larger than 8.0.5, it will got error in compiling stage.\nxtl = \'^0.7.5\'\nxtensor_blas = \'^0.20.0\'\nxtensor_python = \'^0.26.1\'\nxtensor_fftw = \'^0.2.6\' # not show in conda env\nkfr = \'^5.0.2\' # not show in conda env\n\n### [optional-for-mac]\nlibomp = \'^16.0.6\' # via brew',
    'author': 'Fu Yin',
    'author_email': 'oucyinfu@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.12',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



import time
import scipy
import numpy as np

from tqdm import tqdm 
from joblib import Parallel, delayed
from noiseflow.cc.python.utils import moving_ave, split


def xcorr(source_data, receiver_data):
    rfft_corr_data = np.conj(source_data) * receiver_data

    return rfft_corr_data


def deconv(source_data, receiver_data, smoothspect_N):
    win_num = int(source_data.shape[0])
    rfft_npts = int(source_data.shape[1])
    ss_data = np.empty((win_num, rfft_npts), dtype=source_data.dtype)

    for i in range(win_num):
        ss = source_data[i, :]
        temp = moving_ave(np.abs(ss), smoothspect_N)
        ss_data[i] = np.divide(ss, 
                               temp, 
                               out=np.zeros_like(ss, dtype=ss.dtype), 
                               where=temp!=0)

    rfft_corr_data = np.conj(ss_data) * receiver_data

    return rfft_corr_data


def coherency(source_data, receiver_data, smoothspect_N):
    win_num = int(source_data.shape[0])
    rfft_npts = int(source_data.shape[1])

    ss_data = np.empty((win_num, rfft_npts), dtype=source_data.dtype)
    rr_data = np.empty((win_num, rfft_npts), dtype=source_data.dtype)

    for i in range(win_num):
        ss = source_data[i, :]
        rr = receiver_data[i, :]
        ss_temp = moving_ave(np.abs(ss), smoothspect_N)
        rr_temp = moving_ave(np.abs(rr), smoothspect_N)
        ss_data[i] = np.divide(ss, 
                                ss_temp, 
                                out=np.zeros_like(ss, dtype=ss.dtype), 
                                where=ss_temp!=0)
        
        rr_data[i] = np.divide(rr,
                                rr_temp,
                                out=np.zeros_like(rr, dtype=rr.dtype),
                                where=rr_temp!=0)


    rfft_corr_data = np.conj(ss_data) * rr_data

    return rfft_corr_data


class CorrClass_python(object):
    def __init__(self, rfft_data, dt, corr_method, corr_pair, maxlag,
                smoothspect_N, flag, jobs):
        
        self.rfft_data = rfft_data
        self.dt = dt
        self.corr_method = corr_method
        self.corr_pair = corr_pair
        self.maxlag = maxlag
        self.smoothspect_N = smoothspect_N
        self.flag = flag
        self.jobs = jobs

        self.pair_num = int(corr_pair.shape[0])
        self.channel_num = int(rfft_data.shape[0])
        self.win_num = int(rfft_data.shape[1])
        self.rfft_npts = int(rfft_data.shape[2])
        self.npts = int(2*(self.rfft_npts-1))

        t = np.arange(-self.rfft_npts + 1, self.rfft_npts) * dt
        ind = np.where(np.abs(t) <= maxlag)[0]
        self.maxlag_npts = int(ind.size)
        self.maxlag_start = int(ind[0])
        self.maxlag_end = int(ind[-1] + 1)

        # initialize output_data
        if self.rfft_data.dtype == np.dtype(np.complex64):
            self.output_data = np.empty((self.pair_num, self.win_num, self.maxlag_npts), dtype=np.float32)
            self.time_dtype = np.float32
            self.freq_dtype = np.complex64
        elif self.rfft_data.dtype == np.dtype(np.complex128):
            self.output_data = np.empty((self.pair_num, self.win_num, self.maxlag_npts), dtype=np.float64)
            self.time_dtype = np.float64
            self.freq_dtype = np.complex128
        else:
            print("error: please input the correct data type.")
            exit(1)


    def corr(self, source_data, receiver_data):
        if self.corr_method == "xcorr":
            rfft_corr_data = xcorr(source_data, receiver_data)
        elif self.corr_method == "deconv":
            rfft_corr_data = deconv(source_data, receiver_data, self.smoothspect_N)
        elif self.corr_method == "coherency":
            rfft_corr_data = coherency(source_data, receiver_data, self.smoothspect_N)
        else:
            print("error: please input the correct corr method.")
            exit(1)
        
        return rfft_corr_data


    def irfft(self, rfft_corr_data):
        corr_data = np.empty((self.win_num, self.npts), dtype=self.time_dtype)


        for i in range(self.win_num):
            ss = rfft_corr_data[i]
            corr_data[i] = np.roll(scipy.fft.irfft(ss), int(self.rfft_npts - 1))

        return corr_data


    def cut_maxlag(self, corr_data):
        maxlag_corr_data = corr_data[:, self.maxlag_start:self.maxlag_end]

        return maxlag_corr_data


    def process_chunk(self, chunk_start, chunk_end):
        results = []
        if self.flag and chunk_start == 0:
            bar = tqdm(range(chunk_start, chunk_end))
        else:
            bar = range(chunk_start, chunk_end)

        for i in bar:
            # corr
            source_data = self.rfft_data[self.corr_pair[i, 0], :, :]
            receiver_data = self.rfft_data[self.corr_pair[i, 1], :, :]
            rfft_corr_data = self.corr(source_data, receiver_data)

            # irfft
            corr_data = self.irfft(rfft_corr_data)

            # cut maxlag
            maxlag_corr_data = self.cut_maxlag(corr_data)

            results.append(maxlag_corr_data)

        return results


    def run(self):
        if self.flag:
            start_time = time.time()
            print(f"Start corr with {self.jobs} jobs in python...")
        
        # parallel processing
        if self.jobs > 1:
            chunk_list = split(num=self.pair_num, n_jobs=self.jobs)
            results = Parallel(n_jobs=self.jobs, backend="loky")(delayed(self.process_chunk)(chunk_start, chunk_end) for chunk_start, chunk_end in chunk_list)

            for i, chunk_results in enumerate(results):
                chunk_start = chunk_list[i][0]
                for j, result in enumerate(chunk_results):
                    self.output_data[chunk_start + j] = result
        else:
            results = self.process_chunk(0, self.pair_num)
            self.output_data = np.array(results, dtype=self.time_dtype)

        if self.flag:
            end_time = time.time()
            elapsed_time = end_time - start_time
            print(f"End corr with total time {elapsed_time}s")






import os
import re
import json
import numpy as np

from noiseflow.signal.python.taper import taper_py
from noiseflow.signal.python.detrend import detrend_py
from noiseflow.signal.python.decimate import decimate_py
from noiseflow.signal.python.filter import bandpass_py, bandstop_py, lowpass_py, highpass_py

# Determine the absolute path to the config file
env = os.environ.get('CONDA_DEFAULT_ENV')

if env is None:
    env = ''
else:
    env = re.sub('[^a-zA-Z0-9_]', '', env)[0:50]
    
config_path = os.path.abspath(os.path.expanduser(f'~/.noiseflow/config_{env}.json'))    

# Read the config.json file
with open(config_path) as f:
    config = json.load(f)

if config.get('NOISEFLOW_USE_CPP', False):
    NOISEFLOW_USE_CPP = True
    from noiseflow.lib import signal_share
else:
    NOISEFLOW_USE_CPP = False 



def detrend(data, type='linear', flag=False, flag_gap=None, threads=1, py=False):
    # check demision of data
    if data.ndim == 1:
        data = data.reshape(1, -1)
    
    # check type
    if type not in ['linear', 'demean']:
        raise ValueError("type must be 'linear' or 'demean'")
    
    # check threads
    if threads > data.shape[0]:
        raise ValueError("threads must be <= (data.shape[0]=%d)" % data.shape[0])
    
    # check flag_gap
    if flag_gap==None:
        flag_gap = int(data.shape[0])

    if py:
        if data.dtype == np.dtype(np.float32) or data.dtype == np.dtype(np.float64):
            detrend_py(data, type, flag, flag_gap)
        else:
            raise TypeError("data.dtype must be np.float32 or np.float64")
    else:
        if data.dtype == np.dtype(np.float32) and NOISEFLOW_USE_CPP:
            signal_share.detrend_float(data, type, flag, flag_gap, threads)
        elif data.dtype == np.dtype(np.float64) and NOISEFLOW_USE_CPP:
            signal_share.detrend_double(data, type, flag, flag_gap, threads)
        else:
            raise TypeError("data.dtype must be np.float32, np.float64 or NOISEFLOW_USE_CPP=False")
    

    
def taper(data, max_percentage=0.05, type='hann', side='both', flag=False, flag_gap=None, threads=1, py=False):
    # check demision of data
    if data.ndim == 1:
        data = data.reshape(1, -1)

    # check max_percentage
    if max_percentage < 0 or max_percentage > 0.5:
        raise ValueError("max_percentage must be between 0 and 0.5")
    
    # check type
    if type not in ['hann', 'hamming', 'blackman', 'blackman_harris', 'gaussian', 'triangular', 'bartlett', 'cosine', 'cosine_np', 'bartlett_hann', 'bohman', 'lanczos', 'flattop', 'kaiser']:
        raise ValueError("type must be 'hann', 'hamming', 'blackman', 'blackman_harris', 'gaussian', 'triangular', 'bartlett', 'cosine', 'cosine_np', 'bartlett_hann', 'bohman', 'lanczos', 'flattop', 'kaiser'")

    # check side
    if side not in ['both', 'left', 'right']:
        raise ValueError("side must be 'both', 'left', or 'right'")
    
    # check threads
    if threads > data.shape[0]:
        raise ValueError("threads must be <= (data.shape[0]=%d)" % data.shape[0])
    
    # check flag_gap
    if flag_gap==None:
        flag_gap = int(data.shape[0])

    if py:
        if data.dtype == np.dtype(np.float32) or data.dtype == np.dtype(np.float64):
            taper_py(data, max_percentage, type, side, flag, flag_gap)
        else:
            raise TypeError("data.dtype must be np.float32 or np.float64")
    else:
        if data.dtype == np.dtype(np.float32) and NOISEFLOW_USE_CPP:
            signal_share.taper_float(data, max_percentage, type, side, flag, flag_gap, threads)
        elif data.dtype == np.dtype(np.float64) and NOISEFLOW_USE_CPP:
            signal_share.taper_double(data, max_percentage, type, side, flag, flag_gap, threads)
        else:
            raise TypeError("data.dtype must be np.float32, np.float64 or NOISEFLOW_USE_CPP=False")
    


def bandpass(data, freqmin, freqmax, df, corners=4, zerophase=True, flag=False, flag_gap=None, threads=1, py=False):
    # check demision of data
    if data.ndim == 1:
        data = data.reshape(1, -1)

    # check freq and df
    if (freqmin >= df/2) or (freqmax >= df/2):
        raise ValueError("freq must be < df/2")

    # check threads
    if threads > data.shape[0]:
        raise ValueError("threads must be <= (data.shape[0]=%d)" % data.shape[0])
    
    # check flag_gap
    if flag_gap==None:
        flag_gap = int(data.shape[0])

    if py:
        if data.dtype == np.dtype(np.float32) or data.dtype == np.dtype(np.float64):
            bandpass_py(data, freqmin, freqmax, df, corners, zerophase, flag, flag_gap)
        else:
            raise TypeError("data.dtype must be np.float32 or np.float64")
    else:
        if data.dtype == np.dtype(np.float32) and NOISEFLOW_USE_CPP:
            signal_share.bandpass_float(data, freqmin, freqmax, df, corners, zerophase, flag, flag_gap, threads)
        elif data.dtype == np.dtype(np.float64) and NOISEFLOW_USE_CPP:
            signal_share.bandpass_double(data, freqmin, freqmax, df, corners, zerophase, flag, flag_gap, threads)
        else:
            raise TypeError("data.dtype must be np.float32, np.float64 or NOISEFLOW_USE_CPP=False")
    


def bandstop(data, freqmin, freqmax, df, corners=4, zerophase=True, flag=False, flag_gap=None, threads=1, py=False):
    # check demision of data
    if data.ndim == 1:
        data = data.reshape(1, -1)

    # check freq and df
    if (freqmin >= df/2) or (freqmax >= df/2):
        raise ValueError("freq must be < df/2")

    # check threads
    if threads > data.shape[0]:
        raise ValueError("threads must be <= (data.shape[0]=%d)" % data.shape[0])
    
    # check flag_gap
    if flag_gap==None:
        flag_gap = int(data.shape[0])

    if py:
        if data.dtype == np.dtype(np.float32) or data.dtype == np.dtype(np.float64):
            bandstop_py(data, freqmin, freqmax, df, corners, zerophase, flag, flag_gap)
        else:
            raise TypeError("data.dtype must be np.float32 or np.float64")
    else:
        if data.dtype == np.dtype(np.float32) and NOISEFLOW_USE_CPP:
            signal_share.bandstop_float(data, freqmin, freqmax, df, corners, zerophase, flag, flag_gap, threads)
        elif data.dtype == np.dtype(np.float64) and NOISEFLOW_USE_CPP:
            signal_share.bandstop_double(data, freqmin, freqmax, df, corners, zerophase, flag, flag_gap, threads)
        else:
            raise TypeError("data.dtype must be np.float32, np.float64 or NOISEFLOW_USE_CPP=False")
    


def lowpass(data, freq, df, corners=4, zerophase=True, flag=False, flag_gap=None, threads=1, py=False):
    # check demision of data
    if data.ndim == 1:
        data = data.reshape(1, -1)

    # check freq and df
    if freq >= df / 2:
        raise ValueError("freq must be < df/2")

    # check threads
    if threads > data.shape[0]:
        raise ValueError("threads must be <= (data.shape[0]=%d)" % data.shape[0])
    
    # check flag_gap
    if flag_gap==None:
        flag_gap = int(data.shape[0])

    if py:
        if data.dtype == np.dtype(np.float32) or data.dtype == np.dtype(np.float64):
            lowpass_py(data, freq, df, corners, zerophase, flag, flag_gap)
        else:
            raise TypeError("data.dtype must be np.float32 or np.float64")
    else:
        if data.dtype == np.dtype(np.float32) and NOISEFLOW_USE_CPP:
            signal_share.lowpass_float(data, freq, df, corners, zerophase, flag, flag_gap, threads)
        elif data.dtype == np.dtype(np.float64) and NOISEFLOW_USE_CPP:
            signal_share.lowpass_double(data, freq, df, corners, zerophase, flag, flag_gap, threads)
        else:
            raise TypeError("data.dtype must be np.float32, np.float64 or NOISEFLOW_USE_CPP=False")
    


def highpass(data, freq, df, corners=4, zerophase=True, flag=False, flag_gap=None, threads=1, py=False):
    # check demision of data
    if data.ndim == 1:
        data = data.reshape(1, -1)

    # check freq and df
    if freq >= df / 2:
        raise ValueError("freq must be < df/2")

    # check threads
    if threads > data.shape[0]:
        raise ValueError("threads must be <= (data.shape[0]=%d)" % data.shape[0])
    
    # check flag_gap
    if flag_gap==None:
        flag_gap = int(data.shape[0])
    if py:
        if data.dtype == np.dtype(np.float32) or data.dtype == np.dtype(np.float64):
            highpass_py(data, freq, df, corners, zerophase, flag, flag_gap)
        else:
            raise TypeError("data.dtype must be np.float32 or np.float64")
    else:
        if data.dtype == np.dtype(np.float32) and NOISEFLOW_USE_CPP:
            signal_share.highpass_float(data, freq, df, corners, zerophase, flag, flag_gap, threads)
        elif data.dtype == np.dtype(np.float64) and NOISEFLOW_USE_CPP:
            signal_share.highpass_double(data, freq, df, corners, zerophase, flag, flag_gap, threads)
        else:
            raise TypeError("data.dtype must be np.float32, np.float64 or NOISEFLOW_USE_CPP=False")
    


def decimate(data, df, df_new, flag=False, flag_gap=None, threads=1, py=False):
    # check demision of data
    if data.ndim == 1:
        data = data.reshape(1, -1)

    # check df and df_new
    if df % df_new != 0:
        raise ValueError("Note df % df_new == 0")
    if df_new > df:
        raise ValueError("df_new must be <= df")
    
    # check threads
    if threads > data.shape[0]:
        raise ValueError("threads must be <= (data.shape[0]=%d)" % data.shape[0])
    
    # check flag_gap
    if flag_gap==None:
        flag_gap = int(data.shape[0])

    if py:
        if data.dtype == np.dtype(np.float32) or data.dtype == np.dtype(np.float64):
            results = decimate_py(data, df, df_new, flag, flag_gap)
        else:
            raise TypeError("data.dtype must be np.float32 or np.float64")
    else:
        if data.dtype == np.dtype(np.float32) and NOISEFLOW_USE_CPP:
            results = signal_share.decimate_float(data, df, df_new, flag, flag_gap, threads)
        elif data.dtype == np.dtype(np.float64) and NOISEFLOW_USE_CPP:
            results = signal_share.decimate_double(data, df, df_new, flag, flag_gap, threads)
        else:
            raise TypeError("data.dtype must be np.float32, np.float64 or NOISEFLOW_USE_CPP=False")
    
    return results

